if FuBarUtils.VersionToNumber(ace.version) < 103 then
	message("|cffffff7fFuBar|r|cffffffff: Cannot load FuBar since your version of Ace is out of date. Your version: |r|cffffff7f" .. tostring(ace.version) .. "|r|cffffffff, expected version: |r|cffffff7f1.3|r|cffffffff or higher|r")
	return
end

local MAJOR_VERSION, MINOR_VERSION = "1.2", tonumber(string.sub("$Revision: 1647 $", 12, -3))

if FUBAR_REVISION > MINOR_VERSION then
	MINOR_VERSION = FUBAR_REVISION
end
FUBAR_REVISION = nil

local compost = CompostLib:GetInstance('compost-1')
local metro = Metrognome:GetInstance('1')
local dewdrop = DewdropLib:GetInstance('1.0')
local tablet = TabletLib:GetInstance('1.0')
Gratuity:GetInstance('1'):RegisterTooltip(FuBarScanningTooltip)
local jostle = JostleLib:GetInstance('1.0')

FuBar = AceAddon:new({
	name          = FuBarLocals.NAME,
	description   = FuBarLocals.DESCRIPTION,
	version       = MAJOR_VERSION .. "." .. MINOR_VERSION,
	releaseDate   = string.sub("$Date: 2006-05-09 17:18:20 -0500 (Tue, 09 May 2006) $", 8, 17),
	aceCompatible = 103,
	author        = "ckknight",
	email         = "ckknight@gmail.com",
	website       = "http://ckknight.wowinterface.com/",
	category      = "others",
	db            = AceDatabase:new("FuBarDB"),
	defaults      = DEFAULT_OPTIONS,
	cmd           = AceChatCmd:new(FuBarLocals.COMMANDS, FuBarLocals.CMD_OPTIONS),
	loc = FuBarLocals,
})
local FuBar = FuBar

if not FUBAR_DEFAULTS then
	FUBAR_DEFAULTS = function()
		return {
			fontSize = 12,
			adjust = true,
			panels = {
				[1] = {
					attachPoint = "TOP",
					plugins = {
						left = {
							"LocationFu",
							"ExperienceFu",
							"PerformanceFu",
						},
						center = {},
						right = {
							"ClockFu",
							"VolumeFu",
						}
					}
				}, 
				[2] = {
					attachPoint = "BOTTOM"
				}
			}
		}
	end
end

local tmp = private
private = nil
local private = tmp
tmp = nil
private.SPACING_FROM_SIDES = 5
private.MIN_TOOLTIP_SIZE = 200
private.plugins = {}
private.panels = {}
private.highCompatVersions = {}
private.highestCompatVersion = nil
private.lowestCompatVersion = nil

local tmp = PanelClass
PanelClass = nil
local PanelClass = tmp
tmp = nil
local tmp = MinimapContainer
MinimapContainer = nil
local MinimapContainer = tmp
tmp = nil
local FuBarUtils = FuBarUtils

local basePluginFrame

local backgrounds = {["Interface\\AddOns\\FuBar\\background"] = "Default"}

function FuBar:RegisterSkin(name, bgFile)
	assert(type(name) == "string", "You must provide a name for your skin")
	assert(type(bgFile) == "string", "You must provide a file for your skin")
	if not backgrounds[bgFile] then
		backgrounds[bgFile] = name
	end
end

local function SetBackground(bgFile)
	if backgrounds[bgFile] then
		FuBar.data.skin = bgFile
		PanelClass:SetBackground(bgFile)
	end
end

local function background_sort(alpha, bravo)
	return backgrounds[alpha] < backgrounds[bravo]
end

function FuBar:CheckNamingSystem(plugin)
	local wrong = false
	if plugin ~= getglobal(plugin.realName) then
		wrong = true
	else
		local addonExists = false
		for i = 1, GetNumAddOns() do
			local name, title, notes, enabled, loadable, reason, security = GetAddOnInfo(i)
			if name == plugin.realName and IsAddOnLoaded(i) then
				addonExists = true
				local _,_,alpha = string.find(title, "^FuBar %- |cffffffff(.+)|r|cff00ff00Fu|r$")
				local _,_,bravo = string.find(plugin.name, "^FuBar_(.+)Fu$")
				if not alpha or not bravo or alpha ~= bravo then
					_,_,alpha = string.find(title, "^FuBar %- |cffffffff(.+)|r$")
					_,_,bravo = string.find(plugin.name, "^FuBar_(.+)$")
					if not alpha or not bravo or alpha ~= bravo then
						wrong = true
						break
					end
				end
			end
		end
		if not addonExists then
			wrong = true
		end
	end
	if wrong then
		if not plugin.queuedDebugs then
			plugin.queuedDebugs = compost:Acquire()
		end
		table.insert(plugin.queuedDebugs, "You are using nonstandard naming specifications for this plugin. It is strongly recommended you look at README.txt for the proper specifications")
	end
end

function FuBar:GetScaledCursorPosition()
	local x, y = GetCursorPosition()
	local scale = GetScreenHeight() / 768
	return x * scale, y * scale
end

function FuBar:Reset()
	self:UnregisterEvent("PLAYER_LOGOUT")
	FuBarDB = {}
	for _,plugin in ipairs(private.plugins) do
		if plugin.db then
			plugin.db._table = {}
		end
	end
	ReloadUI()
end

function FuBar:GetMinimapContainer()
	return MinimapContainer
end

function FuBar:GetPanel(panelId)
	return private.panels[panelId]
end

function FuBar:GetNumPanels()
	return table.getn(private.panels)
end

function FuBar:GetBottommostTopPanel()
	local bottom = GetScreenHeight()
	local best
	for i = 1, self:GetNumPanels() do
		local panel = self:GetPanel(i)
		if panel:GetAttachPoint() == "TOP" and panel.frame:GetBottom() < bottom then
			bottom = panel.frame:GetBottom()
			best = panel
		end
	end
	return best
end

function FuBar:GetTopmostBottomPanel()
	local top = 0
	local best
	for i = 1, self:GetNumPanels() do
		local panel = self:GetPanel(i)
		if panel:GetAttachPoint() == "BOTTOM" and panel.frame:GetTop() > top then
			top = panel.frame:GetTop()
			best = panel
		end
	end
	return best
end

function FuBar:GetPluginProfiling()
	for _,p in ipairs(private.plugins) do
		if p.realName == self.data.profilePlugin then
			return p, p.realName
		end
	end
end

function FuBar:SetPluginProfiling(plugin)
	self.data.profilePlugin = plugin.realName
end

function FuBar:IsAdjust()
	return self.data.adjust
end

function FuBar:ToggleAdjust(loud)
	self.data.adjust = not self.data.adjust
	if loud then
		self.cmd:status(self.loc.ARGUMENT_ADJUST, self.data.adjust and 1 or 0, self.loc.MAP_ONOFF)
	end
	self:UpdateJostleAdjustments()
	return self.data.adjust
end

function FuBar:IsOverflowing()
	return not self.data.noOverflow
end

function FuBar:ToggleOverflowing()
	self.data.noOverflow = not self.data.noOverflow
	for i = 1, self:GetNumPanels() do
		self:GetPanel(i):CheckForOverlap()
	end
	return not self.data.noOverflow
end

function FuBar:UpdateJostleAdjustments()
	if not self.data.adjust then
		jostle:DisableTopAdjusting()
		jostle:DisableBottomAdjusting()
	else
		if self:IsAutoHidingTop() then
			jostle:DisableTopAdjusting()
		else
			jostle:EnableTopAdjusting()
		end
		if self:IsAutoHidingBottom() then
			jostle:DisableBottomAdjusting()
		else
			jostle:EnableBottomAdjusting()
		end
	end
end

function FuBar:IsAutoHidingTop()
	return self.data.autohideTop
end

function FuBar:ToggleAutoHidingTop(loud)
	self.data.autohideTop = not self.data.autohideTop
	if loud then
		self.cmd:status(self.loc.ARGUMENT_AUTOHIDETOP, self.data.autohideTop and 1 or 0, self.loc.MAP_ONOFF)
	end
	self:UpdateJostleAdjustments()
	if not self.data.autohideTop then
		for i = 1, self:GetNumPanels() do
			if self:GetPanel(i):GetAttachPoint() == "TOP" then
				getglobal("FuBarFrame" .. i):SetAlpha(1)
			end
		end
	else
		metro:Start(self.name .. "AutoHideTop")
	end
	private.autohideTopTime = GetTime()
	return self.data.autohideTop
end

function FuBar:IsAutoHidingBottom()
	return self.data.autohideBottom
end

function FuBar:ToggleAutoHidingBottom(loud)
	self.data.autohideBottom = not self.data.autohideBottom
	if loud then
		self.cmd:status(self.loc.ARGUMENT_AUTOHIDEBOTTOM, self.data.autohideBottom and 1 or 0, self.loc.MAP_ONOFF)
	end
	self:UpdateJostleAdjustments()
	if not self.data.autohideBottom then
		for i = 1, self:GetNumPanels() do
			if self:GetPanel(i):GetAttachPoint() == "BOTTOM" then
				getglobal("FuBarFrame" .. i):SetAlpha(1)
			end
		end
	else
		metro:Start(self.name .. "AutoHideBottom")
	end
	private.autohideBottomTime = GetTime()
	return self.data.autohideBottom
end

function FuBar:GetFontSize()
	if not self.data then
		return 12
	end
	return self.data.fontSize or 12
end
	
function FuBar:SetFontSize(size)
	size = size or 12
	self.data.fontSize = size
	for _,plugin in ipairs(private.plugins) do
		if not plugin.userDefinedFrame then
			if plugin.hasIcon then
				plugin.iconFrame:SetWidth(size + 3)
				plugin.iconFrame:SetHeight(size + 3)
			end
			if not plugin.hasNoText then
				local font, _, flags = plugin.textFrame:GetFont()
				plugin.textFrame:SetFont(font, size, flags)
			end
		else
			plugin:SetFontSize(size)
		end
		plugin:CheckWidth(true)
	end
	for i = 1, self:GetNumPanels() do
		self:GetPanel(i):UpdateTexture()
	end
	self:Update()
	jostle:Refresh()
--	self:AdjustFrames()
end

function FuBar:GetLeftSpacing()
	if not self.data then
		return 20
	end
	return self.data.leftSpacing or 20
end

function FuBar:SetLeftSpacing(size)
	self.data.leftSpacing = size
	self:Update()
end

function FuBar:GetCenterSpacing()
	if not self.data then
		return 20
	end
	return self.data.centerSpacing or 20
end

function FuBar:SetCenterSpacing(size)
	self.data.centerSpacing = size
	self:Update()
end
	
function FuBar:GetRightSpacing()
	if not self.data then
		return 20
	end
	return self.data.rightSpacing or 20
end

function FuBar:SetRightSpacing(size)
	self.data.rightSpacing = size
	self:Update()
end

function FuBar:SetThickness(size)
	self.data.thickness = size
	for i = 1, self:GetNumPanels() do
		self:GetPanel(i):UpdateTexture()
	end
	self:Update()
	jostle:Refresh()
--	self:AdjustFrames()
end

function FuBar:GetThickness(size)
	if not self.data then
		return 5
	end
	return self.data.thickness or 5
end

function FuBar:GetTransparency()
	return self.data.transparency or 0.8
end

function FuBar:SetTransparency(value)
	self.data.transparency = value
	for i = 1, self:GetNumPanels() do
		self:GetPanel(i):UpdateTexture()
	end
end

local function CheckLoadCondition(loadCondition)
	local fuVersion = FuBarUtils.VersionToNumber(FuBar.version)
	return RunScript("(function(fuVersion)"..loadCondition.." end)(" .. fuVersion .. ")")
end

local n
local mt = {
	__newindex = function(self, k, v)
		rawset(self, k, v)
		n[k] = v
	end,
}
local blank = {}
local function LoadAddOnWrapper(addon)
	local o = getfenv(0)
	n = {}
	setmetatable(o, mt)
	local success, ret = pcall(LoadAddOn, addon)
	setmetatable(o, blank)
	if not ret then
		FuBar.cmd:msg("Error loading LoadOnDemand plugin " .. addon)
		return
	elseif not success then
		FuBar.cmd:msg("Error loading LoadOnDemand plugin " .. addon .. ": " .. ret)
		return
	end
	
	local dependencies
	local plugin
	for k,v in pairs(n) do
		local mt = getmetatable(v)
		if mt then
			local good = false
			for _,version in FuBarPlugin.versions do
				if mt.__index == version.instance then
					good = true
					break
				end
			end
			if good then
				if plugin then
					if not dependencies then
						dependencies = {}
					end
					table.insert(dependencies, plugin)
				end
				plugin = v
				plugin.folderName = GetAddOnInfo(addon)
				plugin.realName = k
				if not plugin._initialized then
					plugin._initialized = true
					ace:InitializeApp(plugin)
				end
			end
		end
	end
	if not plugin then
		FuBar.cmd:msg("Error loading LoadOnDemand plugin " .. addon)
		return
	end
	return ret, plugin, dependencies
end

function FuBar:Initialize()
	if not FuBarDB.version or FuBarUtils.VersionToNumber(FuBarDB.version) < 4.0009 then
		FuBarDB = {
			version = FuBarUtils.VersionToNumber(self.version),
			pluginDB = {},
		}
	else
		FuBarDB.version = FuBarUtils.VersionToNumber(self.version)
	end
	
	private.toBeLoaded = {}
	
	if not FuBarDB.profiles then
		FuBarDB.profiles = {}
	end
	
	if not FuBarDB.profiles[self.profilePath[2]] then
		FuBarDB.profiles[self.profilePath[2]] = {
		}
		
		local _,class = UnitClass("player")
		local def
		if self.profilePath[2] == "default" then
			def = FUBAR_DEFAULTS
		else
			def = getglobal("FUBAR_DEFAULTS_" .. class) or FUBAR_DEFAULTS
		end
		if def then
			local function inherit(t, defaults)
				for k,v in pairs(defaults) do
					if type(v) == "table" then
						if t[k] == nil then
							t[k] = {}
						end
						inherit(t[k], v)
					elseif t[k] == nil then
						t[k] = v
					end
				end
			end
			
			if type(def) == "function" then
				def = def()
			end
			if type(def) == "table" then
				inherit(FuBarDB.profiles[self.profilePath[2]], def)
			end
		end
		
		if FuBarDB.profiles[self.profilePath[2]].adjust == nil then
			FuBarDB.profiles[self.profilePath[2]].adjust = true
		end
	end
	
	BossPanel = nil
	BossPanelTooltip = nil
	BossPanelPlugin = nil
	BossPanelUtils = nil
	BossPanelLocals = nil
	DisableAddOn("BossPanel")
	
	self.data = FuBarDB.profiles[self.profilePath[2]]
	
	if not self.data.loadOnDemand then
		self.data.loadOnDemand = {}
	end
	
	if not self.data.tooltip then
		self.data.tooltip = {}
	end
	
	self:SetFontSize(self:GetFontSize())
	
	if self.data.panels then
		local t = compost:Acquire()
		for i,panel in ipairs(self.data.panels) do
			if panel.plugins and panel.plugins.left and panel.plugins.center and panel.plugins.right then
				local left = panel.plugins.left
				local center = panel.plugins.center
				local right = panel.plugins.right
				local u = {}
				for _,p in ipairs(left) do
					if not t[p] then
						t[p] = true
						table.insert(u, p)
					end
				end
				panel.plugins.left = u
				local u = {}
				for _,p in ipairs(center) do
					if not t[p] then
						t[p] = true
						table.insert(u, p)
					end
				end
				panel.plugins.center = u
				local u = {}
				for _,p in ipairs(right) do
					if not t[p] then
						t[p] = true
						table.insert(u, p)
					end
				end
				panel.plugins.right = u
			end
		end
		compost:Reclaim(t)
	end
	
	for i = 1, GetNumAddOns() do
		local name, title, notes, enabled, loadable, reason, security = GetAddOnInfo(i)
		if IsAddOnLoadOnDemand(i) and enabled and loadable and not IsAddOnLoaded(i) then
			local deps = compost:Acquire(GetAddOnDependencies(name))
			for _,dep in ipairs(deps) do
				if dep == "FuBar" then
					if not self.data.loadOnDemand[name] then
						local loaded, plugin, dependencies = LoadAddOnWrapper(name)
						if loaded then
							plugin._loadOnDemand = true
							self.data.loadOnDemand[name] = {
								title = plugin:GetTitle(),
								category = plugin.category,
								disabled = false,
								condition = plugin.loadCondition,
							}
							if dependencies then
								for _,v in ipairs(dependencies) do
									v._loadOnDemand = true
									for j,u in ipairs(private.toBeLoaded) do
										if u == v.realName then
											table.remove(private.toBeLoaded, j)
											break
										end
									end
								end
							end
							self:SetFontSize(self:GetFontSize())
						end
					else
						if not self.data.loadOnDemand[name].disabled and (not self.data.loadOnDemand[name].condition or CheckLoadCondition(self.data.loadOnDemand[name].condition)) then
							local loaded, plugin, dependencies = LoadAddOnWrapper(name)
							if loaded then
								plugin._loadOnDemand = true
								self.data.loadOnDemand[name] = {
									title = plugin:GetTitle(),
									category = plugin.category,
									disabled = false,
									condition = plugin.loadCondition,
								}
								if dependencies then
									for _,v in ipairs(dependencies) do
										v._loadOnDemand = true
										for j,u in ipairs(private.toBeLoaded) do
											if u == v.realName then
												table.remove(private.toBeLoaded, j)
												break
											end
										end
									end
								end
								self:SetFontSize(self:GetFontSize())
							end
						else
							table.insert(private.toBeLoaded, name)
						end
					end
					break
				end
			end
			compost:Reclaim(deps)
		end
	end
	
	self.menu = function(level, value)
		local panelId = private.selectedPanel or 1
		local panel = self:GetPanel(panelId)
		
		if level == 1 then
			dewdrop:AddLine(
				'text', self.name,
				'isTitle', true
			)
			
			local menuTitles = compost:Acquire()
			for _,plugin in ipairs(private.plugins) do
				if not self.loc.MENU_CATEGORIES[plugin.category] then
					plugin.category = "others"
				end
			end
			for _,plugin in ipairs(private.plugins) do
				local alreadyExists = false
				for _,v in ipairs(menuTitles) do
					if v == plugin.category then
						alreadyExists = true
						break
					end
				end
				if not alreadyExists then
					table.insert(menuTitles, plugin.category)
				end
			end
			for _,name in ipairs(private.toBeLoaded) do
				local alreadyExists = false
				for _,v in ipairs(menuTitles) do
					if v == self.data.loadOnDemand[name].category then
						alreadyExists = true
						break
					end
				end
				if not alreadyExists then
					table.insert(menuTitles, self.data.loadOnDemand[name].category)
				end
			end
			table.sort(menuTitles)
			for _,category in ipairs(menuTitles) do
				dewdrop:AddLine(
					'text', self.loc.MENU_CATEGORIES[category],
					'value', category,
					'hasArrow', true
				)
			end
			compost:Reclaim(menuTitles)
			
			dewdrop:AddLine()
			
			dewdrop:AddLine(
				'text', self.loc.MENU_ATTACH,
				'hasArrow', true,
				'value', "attach"
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_LOCK_PANEL,
				'arg1', panel,
				'func', "ToggleLocked",
				'checked', panel:IsLocked()
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_REMOVE_PANEL,
				'arg1', panel,
				'func', "WarnDestroy",
				'disabled', panelId == 1 and not self:GetPanel(2),
				'closeWhenClicked', true
			)
			
			dewdrop:AddLine()
			
			dewdrop:AddLine(
				'text', self.loc.MENU_OVERFLOW_PLUGINS,
				'arg1', self,
				'func', "ToggleOverflowing",
				'checked', self:IsOverflowing()
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_SHOW_NEW_PANEL,
				'arg1', PanelClass,
				'func', "Create"
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_ADJUST_FRAMES,
				'arg1', self,
				'func', "ToggleAdjust",
				'checked', self:IsAdjust()
			)
			
			if panel:GetAttachPoint() ~= "NONE" then
				local hasTop = panel:GetAttachPoint() == "TOP"
				local hasBottom = not hasTop
				if not hasTop then
					for i = 1, self:GetNumPanels() do
						if not hasTop and self:GetPanel(i):GetAttachPoint() == "TOP" then
							hasTop = true
						end
					end
				else
					for i = 1, self:GetNumPanels() do
						if not hasBottom and self:GetPanel(i):GetAttachPoint() == "BOTTOM" then
							hasBottom = true
						end
					end
				end
				if hasTop ~= hasBottom then
					dewdrop:AddLine(
						'text', self.loc.MENU_AUTO_HIDE,
						'arg1', self,
						'func', hasTop and "ToggleAutoHidingTop" or "ToggleAutoHidingBottom",
						'checked', hasTop and self:IsAutoHidingTop() or self:IsAutoHidingBottom()
					)
				elseif panel:GetAttachPoint() == "TOP" then
					dewdrop:AddLine(
						'text', self.loc.MENU_AUTO_HIDE_TOP,
						'arg1', self,
						'func', "ToggleAutoHidingTop",
						'checked', self:IsAutoHidingTop()
					)
				else
					dewdrop:AddLine(
						'text', self.loc.MENU_AUTO_HIDE_BOTTOM,
						'arg1', self,
						'func', "ToggleAutoHidingBottom",
						'checked', self:IsAutoHidingBottom()
					)
				end
			end
			
			dewdrop:AddLine(
				'text', self.loc.MENU_TEXTURE,
				'hasArrow', true,
				'value', "texture"
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_SPACING,
				'hasArrow', true,
				'value', "spacing"
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_FONT_SIZE,
				'hasArrow', true,
				'value', "fontsize"
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_TRANSPARENCY,
				'hasArrow', true,
				'hasSlider', true,
				'sliderFunc', function(value)
					self:SetTransparency(value)
				end,
				'sliderValue', self:GetTransparency()
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_THICKNESS,
				'hasArrow', true,
				'hasSlider', true,
				'sliderFunc', function(value)
					self:SetThickness(value * 20)
					return format("%.0f", value * 20)
				end,
				'sliderValue', self:GetThickness() / 20,
				'sliderBottom', 0,
				'sliderTop', 20
			)
			
			local r, g, b = tablet:GetColor(FuBar.data.tooltip)
			local a = tablet:GetTransparency(FuBar.data.tooltip)
			dewdrop:AddLine(
				'text', self.loc.MENU_TOOLTIP_COLOR,
				'hasColorSwatch', true,
				'hasOpacity', true,
				'r', r,
				'g', g,
				'b', b,
				'opacity', a,
				'swatchFunc', function(r, g, b)
					tablet:SetColor(FuBar.data.tooltip, r, g, b)
				end,
				'opacityFunc', function(a)
					tablet:SetTransparency(FuBar.data.tooltip, a)
				end,
				'cancelFunc', function(r, g, b, a)
					tablet:SetColor(FuBar.data.tooltip, r, g, b)
					tablet:SetTransparency(FuBar.data.tooltip, a)
				end
			)
			
			dewdrop:AddLine(
				'text', self.loc.MENU_PROFILE,
				'hasArrow', true,
				'value', "profile"
			)
			
			dewdrop:AddLine()
			
			dewdrop:AddLine(
				'text', self.loc.MENU_ABOUT,
				'arg1', self.cmd,
				'func', "DisplayAddonInfo"
			)
			
			dewdrop:AddLine(
				'text', CLOSE,
				'arg1', dewdrop,
				'func', "Close"
			)
		elseif level == 2 then
			if value == "profile" then
				dewdrop:AddLine(
					'text', self.loc.MENU_DEFAULT_PROFILE,
					'arg1', AceCommands,
					'func', self.profilePath[2] ~= "default" and "UseProfileDefault",
					'checked', self.profilePath[2] == "default",
					'isRadio', true
				)
				
				dewdrop:AddLine(
					'text', ace.char.class,
					'arg1', AceCommands,
					'func', self.profilePath[2] ~= ace.char.class and "UseProfileClass",
					'arg2', ACE_CMD_PROFILE_ALL,
					'checked', self.profilePath[2] == ace.char.class,
					'isRadio', true
				)
				
				dewdrop:AddLine(
					'text', ace.char.id,
					'arg1', AceCommands,
					'func', self.profilePath[2] ~= ace.char.id and "UseProfileChar",
					'arg2', ACE_CMD_PROFILE_ALL,
					'checked', self.profilePath[2] == ace.char.id,
					'isRadio', true
				)
			elseif value == "spacing" then
				dewdrop:AddLine(
					'text', self.loc.MENU_LEFT_ALIGNED,
					'hasArrow', true,
					'hasSlider', true,
					'sliderFunc', function(value)
						self:SetLeftSpacing(value * 40)
						return format("%.0f", value * 40)
					end,
					'sliderTop', 40,
					'sliderBottom', 0,
					'sliderValue', self:GetLeftSpacing() / 40
				)
				dewdrop:AddLine(
					'text', self.loc.MENU_CENTER_ALIGNED,
					'hasArrow', true,
					'hasSlider', true,
					'sliderFunc', function(value)
						self:SetCenterSpacing(value * 40)
						return format("%.0f", value * 40)
					end,
					'sliderTop', 40,
					'sliderBottom', 0,
					'sliderValue', self:GetCenterSpacing() / 40
				)
				dewdrop:AddLine(
					'text', self.loc.MENU_RIGHT_ALIGNED,
					'hasArrow', true,
					'hasSlider', true,
					'sliderFunc', function(value)
						self:SetRightSpacing(value * 40)
						return format("%.0f", value * 40)
					end,
					'sliderTop', 40,
					'sliderBottom', 0,
					'sliderValue', self:GetRightSpacing() / 40
				)
			elseif value == "texture" then
				local t = compost:Acquire()
				for bgFile in pairs(backgrounds) do
					table.insert(t, bgFile)
				end
				table.sort(t, background_sort)
				for _,bgFile in ipairs(t) do
					dewdrop:AddLine(
						'text', backgrounds[bgFile],
						'func', not PanelClass:IsBackground(bgFile) and SetBackground,
						'arg1', bgFile,
						'checked', PanelClass:IsBackground(bgFile),
						'isRadio', true
					)
				end
				compost:Reclaim(t)
			elseif value == "fontsize" then
				dewdrop:AddLine(
					'text', self.loc.MENU_PANEL,
					'hasArrow', true,
					'hasSlider', true,
					'sliderFunc', function(value)
						value = floor(value * 17 + 7.5)
						self:SetFontSize(value)
						return format("%d", value)
					end,
					'sliderTop', "24",
					'sliderBottom', "7",
					'sliderValue', (self:GetFontSize() - 7) / 17
				)
				local value = tablet:GetFontSizePercent(FuBar.data.tooltip)
				if value < 1 then
					value = value - 0.5
				else
					value = value / 2
				end
				dewdrop:AddLine(
					'text', self.loc.MENU_TOOLTIP,
					'hasArrow', true,
					'hasSlider', true,
					'sliderFunc', function(value)
						if value < 0.5 then
							value = value + 0.5
						else
							value = value * 2
						end
						tablet:SetFontSizePercent(FuBar.data.tooltip, value)
						return format("%.0f%%", value * 100)
					end,
					'sliderTop', "200%",
					'sliderBottom', "50%",
					'sliderValue', value
				)
			elseif value == "attach" then
				dewdrop:AddLine(
					'text', self.loc.MENU_ATTACH_TO_TOP,
					'arg1', panel,
					'func', panel:GetAttachPoint() ~= "TOP" and "SetAttachPoint",
					'arg2', "TOP",
					'checked', panel:GetAttachPoint() == "TOP",
					'closeWhenClicked', true,
					'isRadio', true
				)
				
				dewdrop:AddLine(
					'text', self.loc.MENU_ATTACH_TO_BOTTOM,
					'arg1', panel,
					'func', panel:GetAttachPoint() ~= "BOTTOM" and "SetAttachPoint",
					'arg2', "BOTTOM",
					'checked', panel:GetAttachPoint() == "BOTTOM",
					'closeWhenClicked', true,
					'isRadio', true
				)
				
				dewdrop:AddLine(
					'text', self.loc.MENU_DETACH_PANEL,
					'arg1', panel,
					'func', panel:GetAttachPoint() ~= "NONE" and "SetAttachPoint",
					'arg2', "NONE",
					'checked', panel:GetAttachPoint() == "NONE",
					'closeWhenClicked', true,
					'isRadio', true
				)
			else
				for _,plugin in ipairs(private.plugins) do
					if plugin.category == value then
						dewdrop:AddLine(
							'text', plugin:GetTitle(),
							'arg1', plugin,
							'func', plugin.panel and "Hide" or "Show",
							'arg2', not plugin.panel and private.selectedPanel or 1,
							'checked', plugin.panel
						)
					end
				end
				for i,name in ipairs(private.toBeLoaded) do
					local i, name = i, name
					local plugin = self.data.loadOnDemand[name]
					
					if plugin.category == value then
						dewdrop:AddLine(
							'text', plugin.title,
							'func', function()
								local loaded, plugin, dependencies = LoadAddOnWrapper(name)
								if loaded then
									plugin._loadOnDemand = true
									self.data.loadOnDemand[name] = {
										title = plugin:GetTitle(),
										category = plugin.category,
										disabled = false,
										condition = plugin.loadCondition,
									}
									plugin:Show(private.selectedPanel or 1)
									if dependencies then
										for _,v in ipairs(dependencies) do
											v._loadOnDemand = true
											self.data.loadOnDemand[v.realName].disabled = FALSE
											for j,u in ipairs(private.toBeLoaded) do
												if u == v.realName then
													table.remove(private.toBeLoaded, j)
													break
												end
											end
											if v.hideWithoutStandby then
												v:EnableApp()
											else
												v:Show(panelId)
											end
										end
									end
									self:SetFontSize(self:GetFontSize())
									table.remove(private.toBeLoaded, i)
									self.data.loadOnDemand[name].disabled = FALSE
								end
							end
						)
					end
				end
			end
		end
	end
	local old_ToggleDropDownMenu = ToggleDropDownMenu
	ToggleDropDownMenu = function(alpha, bravo, charlie, delta, echo, foxtrot)
		old_ToggleDropDownMenu(alpha, bravo, charlie, delta, echo, foxtrot)
		if alpha == 1 and delta and type(delta) == "table" and delta.plugin then
			DropDownList1.attachedTo = delta.plugin
		elseif alpha == 1 then
			DropDownList1.attachedTo = nil
		end
	end

	local old_DropDownList1_Hide = DropDownList1.Hide
	DropDownList1.Hide = function(self)
		old_DropDownList1_Hide(self)
		DropDownList1.attachedTo = nil
	end
	
	local tmp = AceCommands.UseProfileDefault
	AceCommands.UseProfileDefault = function(a)
		private.changingProfile = true
		private.previousProfile = self.profilePath[2]
		tmp(a)
		private.changingProfile = false
	end
	
	local tmp = AceCommands.UseProfileClass
	AceCommands.UseProfileClass = function(a, addon)
		private.changingProfile = true
		private.previousProfile = self.profilePath[2]
		tmp(a, addon)
		private.changingProfile = false
	end
	
	local tmp = AceCommands.UseProfileChar
	AceCommands.UseProfileChar = function(a, opt)
		private.changingProfile = true
		private.previousProfile = self.profilePath[2]
		tmp(a, opt)
		private.changingProfile = false
	end
end

function FuBar:Enable()
	self:OnUpdate_UiScale()
	if private.alreadyEnabled then
		ReloadUI()
		return
	end
	private.alreadyEnabled = true
	if not self.data.panels then
		self.data.panels = {}
	end
	
	if not self.data.minimap then
		self.data.minimap = {}
	end
	
	MinimapContainer:Initialize()
	
	for i = table.getn(self.data.panels), 1, -1 do
		if self.data.panels[i] then
			PanelClass:Create()
		end
	end
	
	if not self.data.screenWidth then
		self.data.screenWidth = GetScreenWidth()
	end
	if not self.data.lastUiScale then
		self.data.lastUiScale = 768 / GetScreenHeight()
	end
	
	local screenWidth = GetScreenWidth()
	if self.data.screenWidth ~= screenWidth then
		local ratio = screenWidth / self.data.screenWidth
		for i = 1, self:GetNumPanels() do
			local panel = self:GetPanel(i)
			panel.data.width = panel.data.width * ratio
			panel:Update()
		end
		self.data.screenWidth = screenWidth
	end
	
	local uiScale = 768 / GetScreenHeight()
	if self.data.lastUiScale ~= uiScale then
		local ratio = uiScale / self.data.lastUiScale
		for i = 1, self:GetNumPanels() do
			local panel = self:GetPanel(i)
			panel.data.width = panel.data.width * ratio
			panel:Update()
		end
		self.data.lastUiScale = uiScale
	end
	
	if self:GetNumPanels() == 0 then
		PanelClass:Create()
	end
	
	self:RegisterEvent("ACE_PROFILE_LOADED", "OnAceProfileLoaded")
	self:RegisterEvent("DISPLAY_SIZE_CHANGED", "OnDisplaySizeChanged")
	self:RegisterEvent("CVAR_UPDATE", "OnCVarUpdate")
	self:RegisterEvent("PLAYER_LOGOUT", "OnPlayerLogout")
	
	self:Hook("RestartGx")
	
	local function _OnUpdate(self)
		if not self:SetupPlugins() then
			self._OnUpdate = _OnUpdate
			metro:Start(self.name)
			return
		end
		self:Update()
	end
	self._OnUpdate = _OnUpdate
	
	metro:Register(self.name, self.OnUpdate, 0, self)
	metro:Start(self.name)
	metro:Register(self.name .. "RestartGx", self.OnUpdate_RestartGx, 0, self)
	metro:Register(self.name .. "UiScale", self.OnUpdate_UiScale, 0, self)
	self:OnUpdate_UiScale()
	metro:Register(self.name .. "AutoHideTop", self.OnUpdate_AutoHideTop, 1, self)
	metro:Start(self.name .. "AutoHideTop")
	metro:Register(self.name .. "AutoHideBottom", self.OnUpdate_AutoHideBottom, 1, self)
	metro:Start(self.name .. "AutoHideBottom")
	
	self:UpdateJostleAdjustments()
	
	if self.data.skin and backgrounds[self.data.skin] then
		PanelClass:SetBackground(self.data.skin)
	end
end
	
function FuBar:RestartGx()
	self:CallHook("RestartGx")
	metro:Start(self.name .. "RestartGx")
end

function FuBar:OnCVarUpdate()
	if arg1 == "USE_UISCALE" then
		metro:Start(self.name .. "UiScale")
	end
end

function FuBar:OnUpdate()
	if self._OnUpdate then
		metro:Stop(self.name)
		local f = self._OnUpdate
		self._OnUpdate = nil
		f(self)
	end
end

function FuBar:OnUpdate_RestartGx()
	metro:Stop(self.name .. "RestartGx")
	self:Update()
end

function FuBar:OnUpdate_AutoHideTop()
	metro:Stop(self.name .. "AutoHideTop")
	if not private.inTopPanel and self:IsAutoHidingTop() then
		for i = 1, self:GetNumPanels() do
			if self:GetPanel(i):GetAttachPoint() == "TOP" then
				getglobal("FuBarFrame" .. i):SetAlpha(0)
			end
		end
	end
end

function FuBar:OnUpdate_AutoHideBottom()
	metro:Stop(self.name .. "AutoHideBottom")
	if not private.inBottomPanel and self:IsAutoHidingBottom() then
		for i = 1, self:GetNumPanels() do
			if self:GetPanel(i):GetAttachPoint() == "BOTTOM" then
				getglobal("FuBarFrame" .. i):SetAlpha(0)
			end
		end
	end
end

function FuBar:OnUpdate_UiScale()
	metro:Stop(self.name .. "UiScale")
	local UiScale = 768 / GetScreenHeight()
	if self.data.lastUiScale ~= UiScale then
		local ratio = UiScale / (self.data.lastUiScale or 1)
		for i = 1, self:GetNumPanels() do
			local panel = self:GetPanel(i)
			panel.data.width = panel.data.width / ratio
			panel.data.x = panel.data.x / ratio
			panel:Update()
		end
		self.data.lastUiScale = UiScale
		self.data.screenWidth = GetScreenWidth()
		jostle:Refresh()
	end
end

function FuBar:OnAceProfileLoaded()
	private.changingProfile = false
	if private.previousProfile == self.profilePath[2] then
		return
	end
	if not FuBarDB.profiles[self.profilePath[2]] then
		FuBarDB.profiles[self.profilePath[2]] = {}
		
		local _,class = UnitClass("player")
		local def
		if self.profilePath[2] == "default" then
			def = FUBAR_DEFAULTS
		else
			def = getglobal("FUBAR_DEFAULTS_" .. class) or FUBAR_DEFAULTS
		end
		if def then
			local function inherit(t, defaults)
				for k,v in pairs(defaults) do
					if type(v) == "table" then
						if t[k] == nil then
							t[k] = {}
						end
						inherit(t[k], v)
					elseif t[k] == nil then
						t[k] = v
					end
				end
			end
			
			if type(def) == "function" then
				def = def()
			end
			if type(def) == "table" then
				inherit(FuBarDB.profiles[self.profilePath[2]], def)
			end
		end
		
		if FuBarDB.profiles[self.profilePath[2]].adjust == nil then
			FuBarDB.profiles[self.profilePath[2]].adjust = true
		end
	end
	self.data = FuBarDB.profiles[self.profilePath[2]]
	
	if not self.data.loadOnDemand then
		self.data.loadOnDemand = {}
		local old_data = FuBarDB.profiles[private.previousProfile]
		self.data.loadOnDemand = {}
		for key, value in pairs(old_data.loadOnDemand) do
			if type(value) == "table" then
				if not self.data.loadOnDemand[key] then
					self.data.loadOnDemand[key] = {}
				end
				self.data.loadOnDemand[key].title = old_data.loadOnDemand[key].title
				self.data.loadOnDemand[key].category = old_data.loadOnDemand[key].category
				self.data.loadOnDemand[key].condition = old_data.loadOnDemand[key].condition
			end
		end
	end
	if not self.data.tooltip then
		self.data.tooltip = {}
	end
	
	if not self.data.panels then
		self.data.panels = {}
	end
	
	private.doneSetupPlugins = false
	private.pluginsToSetup = {}
	for i,plugin in ipairs(private.plugins) do
		table.insert(private.pluginsToSetup, plugin)
		plugin.frame:Hide()
		plugin.frame:ClearAllPoints()
		plugin.panel = nil
	end
	
	for i = 1, self:GetNumPanels() do
		local panel = self:GetPanel(i)
		panel.data = nil
		panel.plugins = nil
		panel.frame:Hide()
	end
	private.panels = {}
	private.PanelClass_id = 0
	
	for i = table.getn(self.data.panels), 1, -1 do
		if self.data.panels[i] then
			PanelClass:Create()
		end
	end
	
	if self:GetNumPanels() == 0 then
		PanelClass:Create()
	end
	
	if not self.data.screenWidth then
		self.data.screenWidth = GetScreenWidth()
	end
	if not self.data.lastUiScale then
		self.data.lastUiScale = 768 / GetScreenHeight()
	end
	
	local screenWidth = GetScreenWidth()
	if self.data.screenWidth ~= screenWidth then
		local ratio = screenWidth / self.data.screenWidth
		for i = 1, self:GetNumPanels() do
			local panel = self:GetPanel(i)
			panel.data.width = panel.data.width * ratio
		end
		self.data.screenWidth = screenWidth
	end
	
	local uiScale = 768 / GetScreenHeight()
	if self.data.lastUiScale ~= uiScale then
		local ratio = uiScale / self.data.lastUiScale
		for i = 1, self:GetNumPanels() do
			local panel = self:GetPanel(i)
			panel.data.width = panel.data.width * ratio
		end
		self.data.lastUiScale = uiScale
	end
	
	if not self.data.loadOnDemand then
		self.data.loadOnDemand = {}
	end
	
	for i = 1, GetNumAddOns() do
		local name, title, notes, enabled, loadable, reason, security = GetAddOnInfo(i)
		if IsAddOnLoadOnDemand(i) and enabled and loadable and not IsAddOnLoaded(i) then
			for _,dep in ipairs({GetAddOnDependencies(name)}) do
				if dep == "FuBar" --[[ or dep == "ckknightPanel" ]] then
					if not self.data.loadOnDemand[name] then
						local loaded, plugin, dependencies = LoadAddOnWrapper(name)
						if loaded then
							plugin._loadOnDemand = true
							self.data.loadOnDemand[name] = {
								title = plugin:GetTitle(),
								category = plugin.category,
								disabled = false,
								condition = plugin.loadCondition,
							}
							if dependencies then
								for _,v in ipairs(dependencies) do
									v._loadOnDemand = true
									for j,u in ipairs(private.toBeLoaded) do
										if u == v.realName then
											table.remove(private.toBeLoaded, j)
											break
										end
									end
								end
							end
							self:SetFontSize(self:GetFontSize())
						end
					else
						if not self.data.loadOnDemand[name].disabled and (not self.data.loadOnDemand[name].condition or CheckLoadCondition(self.data.loadOnDemand[name].condition)) then
							local loaded, plugin, dependencies = LoadAddOnWrapper(name)
							if loaded then
								plugin._loadOnDemand = true
								self.data.loadOnDemand[name] = {
									title = plugin:GetTitle(),
									category = plugin.category,
									disabled = false,
									condition = plugin.loadCondition,
								}
								if dependencies then
									for _,v in ipairs(dependencies) do
										v._loadOnDemand = true
										for j,u in ipairs(private.toBeLoaded) do
											if u == v.realName then
												table.remove(private.toBeLoaded, j)
												break
											end
										end
									end
								end
								self:SetFontSize(self:GetFontSize())
							end
						else
							local exists = false
							for _,n in ipairs(private.toBeLoaded) do
								if n == name then
									exists = true
									break
								end
							end
							if not exists then
								table.insert(private.toBeLoaded, name)
							end
						end
					end
					break
				end
			end
		end
	end

	self:SetupPlugins()
	
	self:SetFontSize(self:GetFontSize())
	
	local function _OnUpdate(self)
		if not self:SetupPlugins() then
			self._OnUpdate = _OnUpdate
			metro:Start(self.name)
			return
		end
		self:Update()
	end
	self._OnUpdate = _OnUpdate
	metro:Start(self.name)
end

function FuBar:Disable()
	for _,plugin in ipairs(private.plugins) do
		plugin:DisableAddon()
	end
	for i = 1, self:GetNumPanels() do
		self:GetPanel(i).frame:Hide()
	end
--	self:AdjustFrames()
end

function FuBar:Report()
	local report = compost:Acquire(
		compost:AcquireHash(
			'text', self.loc.ARGUMENT_ADJUST,
			'val', self:IsAdjust() and 1 or 0,
			'map', self.loc.MAP_ONOFF
		)
	)
	self.cmd:report(report)
	compost:Reclaim(report, 1)
end

function FuBar:GetNumPlugins()
	return table.getn(private.plugins)
end

function FuBar:GetPlugin(i)
	return private.plugins[i]
end

function FuBar:AcquirePluginFrame(plugin)
	local id = self:GetNumPlugins() + 1
	local name = "FuBarPluginFrame" .. id
	local frame = getglobal(name)
	if not frame then
		frame = self:CreateBasicPluginFrame(name, plugin)
		
		local icon = frame:CreateTexture(name .. "Icon", "ARTWORK")
		icon:SetWidth(16)
		icon:SetHeight(16)
		icon:SetPoint("LEFT", frame, "LEFT")
		
		local text = frame:CreateFontString(name .. "Text", "ARTWORK")
		text:SetWidth(134)
		text:SetHeight(24)
		text:SetPoint("LEFT", icon, "RIGHT", 0, 1)
		text:SetFontObject(GameFontNormal)
	end
	return frame
end

function FuBar:CreateBasicPluginFrame(name, plugin)
	local frame = CreateFrame("Button", name, UIParent)
	frame:SetFrameStrata("HIGH")
	frame:EnableMouse(true)
	frame:EnableMouseWheel(true)
	frame:SetMovable(true)
	frame:SetWidth(150)
	frame:SetHeight(24)
	frame:SetPoint("CENTER", UIParent, "CENTER")
	
	local downTime
	frame:SetScript("OnClick", function()
		if (not downTime or GetTime() < downTime + 0.5) and (not this.stopClick or GetTime() > this.stopClick) then
			FuBar:Plugin_OnClick(this.plugin)
		end
		downTime = nil
	end)
	frame:SetScript("OnDoubleClick", function()
		FuBar:Plugin_OnDoubleClick(this.plugin)
	end)
	frame:SetScript("OnMouseDown", function()
		if arg1 == "LeftButton" and not IsShiftKeyDown() and not IsControlKeyDown() and not IsAltKeyDown() then
			HideDropDownMenu(1)
			downTime = GetTime()
			FuBar:Plugin_StartDrag(this.plugin)
			FuBar:Plugin_OnMouseDown(this.plugin, arg1)
		elseif arg1 == "RightButton" and not IsShiftKeyDown() and not IsControlKeyDown() and not IsAltKeyDown() then
			FuBar:Plugin_OpenMenu(this.plugin)
		else
			HideDropDownMenu(1)
			FuBar:Plugin_OnMouseDown(this.plugin, arg1)
		end
	end)
	frame:SetScript("OnMouseUp", function()
		if arg1 == "LeftButton" then
			if not FuBar:Plugin_StopDrag(this.plugin) then
				FuBar:Plugin_OnMouseUp(this.plugin, arg1)
			else
				this.stopClick = GetTime() + 0.05
			end
		else
			FuBar:Plugin_OnMouseUp(this.plugin, arg1)
		end
	end)
	frame:SetScript("OnEnter", function()
		FuBar:Panel_OnEnter(this.plugin)
	end)
	frame:SetScript("OnLeave", function()
		FuBar:Panel_OnLeave(this.plugin)
	end)
	return frame
end

local function inheritDefaults(t, defaults)
	if not defaults then
		return
	end
	for k,v in pairs(defaults) do
		if type(v) == "table" then
			if type(t[k]) ~= "table" then
				t[k] = {}
			end
			inheritDefaults(t[k], v)
		elseif t[k] == nil then
			t[k] = v
		end
	end
end

local function cleanDefaults(t, defaults, notFirst)
	if defaults then
		for k,v in pairs(defaults) do
			if type(v) == "table" then
				if type(t[k]) == "table" then
					cleanDefaults(t[k], v, true)
					if next(t[k]) == nil then
						t[k] = nil
					end
				end
			elseif t[k] == v then
				t[k] = nil
			elseif t[k] == true and v == TRUE then
				t[k] = nil
			elseif t[k] == TRUE and v == true then
				t[k] = nil
			end
		end
		for k,v in pairs(t) do
			if v == false and not defaults[k] then
				t[k] = nil
			end
		end
	end
	if not notFirst then
		if next(t) == "version" and next(t, "version") == nil then
			t.version = nil
			return true
		end
	end
	return false
end

local tablet = TabletLib:GetInstance('1.0')
function FuBar:RegisterPlugin(plugin)
	self:CheckNamingSystem(plugin)
	
	table.insert(private.plugins, plugin)
	
end

function FuBar:Update()
	for panelId = 1, self:GetNumPanels() do
		self:GetPanel(panelId):Update()
	end
end

function FuBar:SetupPlugins()
	private.doneSetupPlugins = true
	if private.pluginsToSetup then
		for i = 1, self:GetNumPanels() do
			local panel = self:GetPanel(i)
			for h = 1, 3 do
				local side
				if h == 1 then
					side = "LEFT"
				elseif h == 2 then
					side = "CENTER"
				else
					side = "RIGHT"
				end
				local order = panel:GetSavedOrder(side)
				for _,name in ipairs(order) do
					for i,plugin in ipairs(private.pluginsToSetup) do
						if not plugin.disabled and plugin.realName == name then
							if not panel:AddPlugin(plugin, nil, side) then
								private.doneSetupPlugins = false
								return
							end
							table.remove(private.pluginsToSetup, i)
							break
						end
					end
				end
			end
		end
		
		local order = MinimapContainer:GetSavedOrder()
		for _,name in ipairs(order) do
			for i,plugin in ipairs(private.pluginsToSetup) do
				if not plugin.disabled and plugin.realName == name then
					if not MinimapContainer:AddPlugin(plugin) then
						private.doneSetupPlugins = false
						return
					end
					table.remove(private.pluginsToSetup, i)
					break
				end
			end
		end
		
		for _,plugin in ipairs(private.pluginsToSetup) do
			if plugin.defaultPosition == "MINIMAP" then
				MinimapContainer:AddPlugin(plugin)
			else
				self.lastPanelId = (self.lastPanelId or 0) + 1
				if self.lastPanelId > self:GetNumPanels() then
					self.lastPanelId = 1
				end
				self:GetPanel(self.lastPanelId):AddPlugin(plugin, nil, plugin.defaultPosition or "LEFT")
			end
		end
		
		private.pluginsToSetup = nil
	end
	return true
end

function FuBar:ShowPlugin(plugin, panelId)
	if private.doneSetupPlugins then
		if not panelId then
			if plugin.defaultPosition == "MINIMAP" then
				MinimapContainer:AddPlugin(plugin)
				return
			end
			self.lastPanelId = (self.lastPanelId or 0) + 1
			if self.lastPanelId > self:GetNumPanels() then
				self.lastPanelId = 1
			end
			panelId = self.lastPanelId
		end
		self:GetPanel(panelId):AddPlugin(plugin, nil, plugin.defaultPosition or "LEFT")
	else
		if not private.pluginsToSetup then
			private.pluginsToSetup = {}
		end
		table.insert(private.pluginsToSetup, plugin)
	end
end

function FuBar:OnDisplaySizeChanged()
	if not self.data.screenWidth then
		self.data.screenWidth = GetScreenWidth()
	end
	if not self.data.lastUiScale then
		self.data.lastUiScale = 768 / GetScreenHeight()
	end
	
	local screenWidth = GetScreenWidth()
	if self.data.screenWidth ~= screenWidth then
		local ratio = screenWidth / self.data.screenWidth
		for i = 1, self:GetNumPanels() do
			local panel = self:GetPanel(i)
			panel.data.width = panel.data.width * ratio
			panel:Update()
		end
		self.data.screenWidth = screenWidth
	end
	
	local uiScale = 768 / GetScreenHeight()
	if self.data.lastUiScale ~= uiScale then
		local ratio = uiScale / self.data.lastUiScale
		for i = 1, self:GetNumPanels() do
			local panel = self:GetPanel(i)
			panel.data.width = panel.data.width * ratio
			panel:Update()
		end
		self.data.lastUiScale = uiScale
	end
	self:Update()
end

function FuBar:Plugin_OpenMenu(plugin)
	if not plugin.panel then
		return
	end
	if dewdrop:IsOpen(plugin.frame) then
		dewdrop:Close()
		return
	end
	tablet:Close()
	
	if not dewdrop:IsRegistered(plugin.frame) then
		dewdrop:Register(plugin.frame, 
			'children', function(level, value, valueN_1, valueN_2, valueN_3, valueN_4)
				if level == 1 then
					dewdrop:AddLine(
						'text', plugin:GetTitle(),
						'isTitle', true
					)
				end
				
				if plugin.MenuSettings then
					plugin:MenuSettings(level, value, false, valueN_1, valueN_2, valueN_3, valueN_4)
				end
				
				if level == 1 and not plugin.overrideMenu then
					if plugin.MenuSettings then
						dewdrop:AddLine()
					end
					plugin:AddImpliedMenuOptions()
				end
				if level == 1 then
					dewdrop:AddLine(
						'text', CLOSE,
						'func', dewdrop.Close,
						'arg1', dewdrop
					)
				end
			end,
			'point', function()
				local x, y = plugin.frame:GetCenter()
				local leftRight
				if x < GetScreenWidth() / 2 then
					leftRight = "LEFT"
				else
					leftRight = "RIGHT"
				end
				if y < GetScreenHeight() / 2 then
					return "BOTTOM" .. leftRight, "TOP" .. leftRight
				else
					return "TOP" .. leftRight, "BOTTOM" .. leftRight
				end
			end,
			'dontHook', true
		)
	end
	if not plugin.minimapFrame or not plugin:IsMinimapAttached() then
		dewdrop:Open(plugin.frame)
	else
		dewdrop:Open(plugin.minimapFrame, plugin.frame)
	end
end

function FuBar:Plugin_StartDrag(plugin)
	self:Panel_OnLeave()
	local panel = plugin.panel
	local index, side = panel:IndexOfPlugin(plugin)
	
	if index then
		if not panel:IsLocked() then
			if index < panel:GetNumPlugins(side) then
				local frame = panel:GetPlugin(index + 1, side).frame
				local x, y = frame:GetLeft(), frame:GetBottom()
				frame:ClearAllPoints()
				frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", x, y)
			end
			
			private.previousPosition_panel = panel
			private.previousPosition_index = index
			private.previousPosition_side = side
			private.previousPosition_x, private.previousPosition_y = self:GetScaledCursorPosition()
			
			private.PanelClass_stopUpdates = true
			panel:RemovePlugin(index, side)
			plugin.frame:StartMoving()
		end
	end
end

function FuBar:Plugin_StopDrag(plugin)
	plugin.frame:StopMovingOrSizing()
	if private.previousPosition_panel and not private.previousPosition_panel:IsLocked() then
		local insertAt
		local panel = private.previousPosition_panel
		local side = private.previousPosition_side
		local x, y = self:GetScaledCursorPosition()
		local px, py = private.previousPosition_x, private.previousPosition_y

		for panelId = 1, self:GetNumPanels() do
			local p = self:GetPanel(panelId)
			local frame = p.frame
			local x1, x2, y1, y2 = frame:GetLeft(), frame:GetRight(), frame:GetBottom(), frame:GetTop()
			if x1 <= x and x <= x2 and y1 <= y and y <= y2 then
				panel = p
				break
			end
		end
		
		if panel:GetNumPlugins(side) == 0 then
			insertAt = 1
		else
			for i = 1, panel:GetNumPlugins(side) + 1 do
				local left, right
				local leftPlugin
				local rightPlugin
				
				leftPlugin = panel:GetPlugin(i - 1, side)
				rightPlugin = panel:GetPlugin(i, side)
				
				if side == "RIGHT" then
					leftPlugin, rightPlugin = rightPlugin, leftPlugin
				end
				left = leftPlugin and leftPlugin.frame:GetCenter() or 0
				right = rightPlugin and rightPlugin.frame:GetCenter() or GetScreenWidth()
				if left <= x and x <= right then
					insertAt = i
					break
				end
			end
		end
		if not insertAt then
			insertAt = panel:GetNumPlugins(side) + 1
		end
		
		panel:AddPlugin(plugin, insertAt, side)
		private.PanelClass_stopUpdates = false
		private.previousPosition_panel:StopPreventUpdate()
		private.previousPosition_panel:Update()
		panel:Update()
		private.previousPosition_panel = nil
		private.previousPosition_index = nil
		private.previousPosition_side = nil
		private.previousPosition_x = nil
		private.previousPosition_y = nil
		if math.abs(px - x) <= 5 and math.abs(py - y) <= 5 then
			return false
		else
			return true
		end
	end
end

function FuBar:Plugin_OnClick(plugin)
	if not plugin.disabled and plugin.OnClick then
		plugin:OnClick()
	end
end

function FuBar:Plugin_OnDoubleClick(plugin)
	if not plugin.disabled and plugin.OnDoubleClick then
		plugin:OnDoubleClick()
	end
end

function FuBar:Plugin_OnMouseDown(plugin, button)
	if not plugin.disabled and plugin.OnMouseDown then
		plugin:OnMouseDown(button)
	end
end

function FuBar:Plugin_OnMouseUp(plugin, button)
	if not plugin.disabled and plugin.OnMouseUp then
		plugin:OnMouseUp(button)
	end
end

function FuBar:Panel_OnEnter(plugin)
	local point = plugin
	if type(plugin) == "table" then
		point = plugin.panel:GetAttachPoint()
	end
	if point == "TOP" then
		private.inTopPanel = true
		private.autohideTopTime = nil
		metro:Stop(self.name .. "AutoHideTop")
	elseif point == "BOTTOM" then
		private.inBottomPanel = true
		private.autohideBottomTime = nil
		metro:Stop(self.name .. "AutoHideBottom")
	end
	for i = 1, self:GetNumPanels() do
		local panel = self:GetPanel(i)
		if panel:GetAttachPoint() == point then
			panel.frame:SetAlpha(1)
		end
	end
end

function FuBar:Panel_OnLeave(plugin)
	local point = plugin
	if type(plugin) == "table" then
		point = plugin.panel:GetAttachPoint()
	end
	if point == "TOP" then
		private.inTopPanel = false
		private.autohideTopTime = GetTime()
		metro:Start(self.name .. "AutoHideTop")
	elseif point == "BOTTOM" then
		private.inBottomPanel = false
		private.autohideBottomTime = GetTime()
		metro:Start(self.name .. "AutoHideBottom")
	end
end

function FuBar:OnPlayerLogout()
	for _,plugin in ipairs(private.plugins) do
		local t
		if plugin.db then
			t = plugin.db._table
		else
			t = FuBarDB.pluginDB[plugin.realName]
		end
		if cleanDefaults(plugin.data, plugin.defaults) then
			if plugin.db then
				t.profiles[plugin.profilePath[2]] = nil
			else
				t.profiles[FuBar.profilePath[2]] = nil
			end
		end
		if type(plugin.chardb) ~= "string" then
			if cleanDefaults(plugin.charData, plugin.charDefaults) then
				t.chars[ace.char.realm .. ace.char.name] = nil
				if next(t.chars) == nil then
					t.chars = nil
				end
			end
		else
			if cleanDefaults(plugin.charData, plugin.charDefaults) then
				setglobal(plugin.chardb, nil)
			end
		end
		if cleanDefaults(plugin.fullData, plugin.fullDefaults) then
			t.full = nil
		end
		if cleanDefaults(plugin.realmData, plugin.realmDefaults) then
			t.realms[ace.char.realm .. ace.char.faction] = nil
			if next(t.realms) == nil then
				t.realms = nil
			end
		end
		if cleanDefaults(plugin.classData, plugin.classDefaults) then
			t.classes[ace.char.class] = nil
			if next(t.classes) == nil then
				t.classes = nil
			end
		end
	end
end

FuBar:RegisterForLoad()
