local MAJOR_VERSION = "1.2"
local MINOR_VERSION = tonumber(string.sub("$Revision: 1646 $", 12, -3))

if FuBarPlugin and FuBarPlugin.versions[MAJOR_VERSION] and FuBarPlugin.versions[MAJOR_VERSION].minor >= MINOR_VERSION then
	return
end

if MINOR_VERSION > FUBAR_REVISION then
	FUBAR_REVISION = MINOR_VERSION
end

-------------IRIEL'S-STUB-CODE--------------
local stub = {};

-- Instance replacement method, replace contents of old with that of new
function stub:ReplaceInstance(old, new)
   for k,v in pairs(old) do old[k]=nil; end
   for k,v in pairs(new) do old[k]=v; end
end

-- Get a new copy of the stub
function stub:NewStub()
  local newStub = {};
  self:ReplaceInstance(newStub, self);
  newStub.lastVersion = '';
  newStub.versions = {};
  return newStub;
end

-- Get instance version
function stub:GetInstance(version)
   if (not version) then version = self.lastVersion; end
   local versionData = self.versions[version];
   if (not versionData) then
      message("Cannot find library instance with version '" 
              .. version .. "'");
      return;
   end
   return versionData.instance;
end

-- Register new instance
function stub:Register(newInstance)
   local version,minor = newInstance:GetLibraryVersion();
   self.lastVersion = version;
   local versionData = self.versions[version];
   if (not versionData) then
      -- This one is new!
      versionData = { instance = newInstance,
         minor = minor,
         old = {} 
      };
      self.versions[version] = versionData;
      newInstance:LibActivate(self);
      return newInstance;
   end
   if (minor <= versionData.minor) then
      -- This one is already obsolete
      if (newInstance.LibDiscard) then
         newInstance:LibDiscard();
      end
      return versionData.instance;
   end
   -- This is an update
   local oldInstance = versionData.instance;
   local oldList = versionData.old;
   versionData.instance = newInstance;
   versionData.minor = minor;
   local skipCopy = newInstance:LibActivate(self, oldInstance, oldList);
   table.insert(oldList, oldInstance);
   if (not skipCopy) then
      for i, old in ipairs(oldList) do
         self:ReplaceInstance(old, newInstance);
      end
   end
   return newInstance;
end

-- Bind stub to global scope if it's not already there
if (not FuBarPlugin) then
   FuBarPlugin = stub:NewStub();
end

-- Nil stub for garbage collection
stub = nil;
-----------END-IRIEL'S-STUB-CODE------------

local function assert(condition, message)
	if not condition then
		local stack = debugstack()
		local first = string.gsub(stack, "\n.*", "")
		local file = string.gsub(first, "^(.*\\.*)%.lua:%d+: .*", "%1")
		if not message then
			local _,_,second = string.find(stack, "\n(.-)\n")
			message = "assertion failed! " .. second
		end
		local i = 1
		for s in string.gfind(stack, "\n(.-)\n") do
			i = i + 1
			if not string.find(s, file .. "%.lua:%d+:") then
				error(message, i)
				return
			end
		end
		error(message, 2)
		return
	end
	return condition
end

local compost = CompostLib:GetInstance('compost-1')
local metro = Metrognome:GetInstance('1')

private = {}
local private = private
local epsilon = 1e-5

local function inheritDefaults(t, defaults)
	if not defaults then
		return
	end
	for k,v in pairs(defaults) do
		if type(v) == "table" then
			if type(t[k]) ~= "table" then
				t[k] = {}
			end
			inheritDefaults(t[k], v)
		elseif t[k] == v then
			t[k] = nil
		end
	end
	for k,v in pairs(t) do
		if v == TRUE and defaults[k] == true then
			t[k] = nil
		elseif v == true and defaults[k] == TRUE then
			t[k] = nil
		end
	end
	setmetatable(t, { __index = defaults })
end
local onClickTypes = {"OnClick", "OnDoubleClick", "OnMouseDown", "OnMouseUp"}

local lib = AceAddon:new()

function lib:GetTitle()
	local _,_,title = string.find(self.name, "FuBar %- (.+)")
	if not title then
		title = self.name
	end
	return title
end

function lib:DisableHideWithoutStandby()
	self.hideWithoutStandby = false
	if not self._hideWithoutStandbyEnabled then
		return
	end
	self._hideWithoutStandbyEnabled = false
	for index, value in self.cmd.options do
		if value.method == "_ToggleVisible" then
			table.remove(self.cmd.options, index)
			break
		end
	end
	
	self._ToggleVisible = nil
	
	self.Report = self._Report
	self._Report = nil
	
	if self.data.hidden then
		self:DisableAddon()
	end
	self.data.hidden = nil
end

function lib:EnableHideWithoutStandby()
	self.hideWithoutStandby = true
	if self._hideWithoutStandbyEnabled then 
		return
	end
	self._hideWithoutStandbyEnabled = true
	table.insert(self.cmd.options, {
		option = FuBarLocals.ARGUMENT_VISIBLE,
		desc = FuBarLocals.TEXT_TOGGLE_VISIBILITY,
		method = "_ToggleVisible"
	})
	
	function self:_ToggleVisible()
		if self.data.hidden then
			self:Show()
		else
			self:Hide()
		end
		self.cmd:status(FuBar_PerformanceLocals.ARGUMENT_VISIBLE, not self.data.hidden and 1 or 0, FuBarLocals.MAP_ONOFF)
	end
	
	if self.Report then
		self._Report = self.Report
		function self:Report()
			self:_Report()
			ace:print(" - ", format(ACE_CMD_REPORT_LINE, FuBarLocals.ARGUMENT_VISIBLE, FuBarLocals.MAP_ONOFF[not self.data.hidden and 1 or 0]))
		end
	else
		function self:Report()
			local report = compost:Acquire(
				compost:AcquireHash(
					'text', FuBarLocals.ARGUMENT_VISIBLE,
					'val', not self.data.hidden and 1 or 0,
					'map', FuBarLocals.MAP_ONOFF
				)
			)
			self.cmd:report(report)
			compost:Reclaim(report, 1)
		end
	end
end

function lib:IsTextColored()
	if self.data then
		return not self.data.uncolored
	else
		return true
	end
end

function lib:ToggleMinimapAttached()
	local value = self:IsMinimapAttached()
	if value then
		self.panel:RemovePlugin(self)
		FuBar:GetPanel(1):AddPlugin(self)
	else
		self.panel:RemovePlugin(self)
		FuBar:GetMinimapContainer():AddPlugin(self)
	end
end

function lib:IsMinimapAttached()
	return self.panel == FuBar:GetMinimapContainer()
end

function lib:ToggleTextColored(loud)
	self.data.uncolored = not self.data.uncolored or nil
	if loud then
		self.cmd:status(FuBarLocals.ARGUMENT_COLOR, not self.data.uncolored and 1 or 0, FuBarLocals.MAP_ONOFF)
	end
	self:UpdateText()
	return not self.data.uncolored
end

function lib:Update()
	self:UpdateData()
	self:UpdateText()
	self:UpdateTooltip()
end

function lib:UpdateDisplay()
	self:UpdateText()
	self:UpdateTooltip()
end

function lib:_OnAceProfileLoaded()
	if self.db then
		local t = self.db._table
		if not t.profiles then
			t.profiles = {}
		end
		if not t.profiles[self.profilePath[2]] then
			t.profiles[self.profilePath[2]] = {}
		end
		if self.data ~= self.db._table.profiles[self.profilePath[2]] then
			self.data = self.db._table.profiles[self.profilePath[2]]
		else
			return
		end
	else
		if not FuBarDB.pluginDB[self.realName] then
			FuBarDB.pluginDB[self.realName] = {}
		end
		if not FuBarDB.pluginDB[self.realName].profiles then
			FuBarDB.pluginDB[self.realName].profiles = {}
		end
		if not FuBarDB.pluginDB[self.realName].profiles[FuBar.profilePath[2]] then
			FuBarDB.pluginDB[self.realName].profiles[FuBar.profilePath[2]] = {}
		end
		if self.data ~= FuBarDB.pluginDB[self.realName].profiles[FuBar.profilePath[2]] then
			self.data = FuBarDB.pluginDB[self.realName].profiles[FuBar.profilePath[2]]
		else
			return
		end
	end
	inheritDefaults(self.data, self.defaults)
	
	if self.ACE_PROFILE_LOADED then
		self:ACE_PROFILE_LOADED()
	end
		
	self.data.version = self.version
	
	self:Update()
end

function lib:Show(panelId)
	if self._loadOnDemand then
		if not FuBar.data.loadOnDemand then
			FuBar.data.loadOnDemand = {}
		end
		if not FuBar.data.loadOnDemand[self.folderName] then
			FuBar.data.loadOnDemand[self.folderName] = {}
		end
		FuBar.data.loadOnDemand[self.folderName].disabled = nil
	end
	if panelId then
		self.data.hidden = nil
	end
	if self.disabled then
		self.panelIdTmp = panelId
		self:EnableAddon()
		self.panelIdTmp = nil
		if not self.db then
			self.data.disabled = nil
		end
	else
		FuBar:ShowPlugin(self, panelId or self.panelIdTmp)
		self:Update()
	end
end

function lib:Hide()
	if self.hideWithoutStandby then
		self.data.hidden = true
	end
	if not self.hideWithoutStandby then
		if not self.overrideTooltip and not self.cannotDetachTooltip and self:IsTooltipDetached() and self.data.detachedTooltip.detached then
			self:ReattachTooltip()
			self.data.detachedTooltip.detached = true
		end
		if not self.db then
			self.data.disabled = true
		end
		self:DisableAddon()
		if self._loadOnDemand then
			FuBar.data.loadOnDemand[self.folderName].disabled = true
		end
	end
	if self.panel then
		self.panel:RemovePlugin(self)
	end
	self.frame:Hide()
end

local dewdrop = DewdropLib:GetInstance('1.0')
local tablet = TabletLib:GetInstance('1.0')
function lib:AddImpliedMenuOptions(level)
	if self.hasIcon and not self.hasNoText then
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_SHOW_ICON,
			'func', function()
				self:ToggleIconShown()
			end,
			'checked', self:IsIconShown(),
			'level', level
		)
	end
	
	if not self.cannotHideText and self.hasIcon and not self.hasNoText then
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_SHOW_TEXT,
			'func', function()
				self:ToggleTextShown()
			end,
			'checked', self:IsTextShown(),
			'leve', level
		)
	end
	
	if not self.userDefinedFrame and not self.hasNoText and not self.hasNoColor then
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_SHOW_COLORED_TEXT,
			'arg1', self,
			'func', "ToggleTextColored",
			'checked', self:IsTextColored(),
			'level', level
		)
	end
	
	if not self.overrideTooltip and not self.cannotDetachTooltip then
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_DETACH_TOOLTIP,
			'func', function()
				if self:IsTooltipDetached() then
					self:ReattachTooltip()
				else
					self:DetachTooltip()
				end
			end,
			'checked', self:IsTooltipDetached(),
			'level', level
		)
		
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_LOCK_TOOLTIP,
			'arg1', tablet,
			'func', "ToggleLocked",
			'arg2', self.frame,
			'checked', tablet:IsLocked(self.frame),
			'disabled', not self:IsTooltipDetached(),
			'level', level
		)
	end
	
	if not self:IsMinimapAttached() then
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_ON_LEFT_SIDE,
			'arg1', self.panel,
			'func', self.panel:GetPluginSide(self) ~= "LEFT" and "SetPluginSide",
			'arg2', self,
			'arg3', "LEFT",
			'checked', self.panel:GetPluginSide(self) == "LEFT",
			'level', level,
			'closeWhenClicked', self.panel:GetPluginSide(self) ~= "LEFT",
			'isRadio', true
		)
		
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_IN_CENTER,
			'arg1', self.panel,
			'func', self.panel:GetPluginSide(self) ~= "CENTER" and "SetPluginSide",
			'arg2', self,
			'arg3', "CENTER",
			'checked', self.panel:GetPluginSide(self) == "CENTER",
			'level', level,
			'closeWhenClicked', self.panel:GetPluginSide(self) ~= "CENTER",
			'isRadio', true
		)
		
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_ON_RIGHT_SIDE,
			'arg1', self.panel,
			'func', self.panel:GetPluginSide(self) ~= "RIGHT" and "SetPluginSide",
			'arg2', self,
			'arg3', "RIGHT",
			'checked', self.panel:GetPluginSide(self) == "RIGHT",
			'level', level,
			'closeWhenClicked', self.panel:GetPluginSide(self) ~= "RIGHT",
			'isRadio', true
		)
	end
	
	dewdrop:AddLine(
		'text', "Attach to minimap",
		'arg1', self,
		'func', "ToggleMinimapAttached",
		'checked', self:IsMinimapAttached(),
		'level', level,
		'closeWhenClicked', true
	)
	
	if self.hideWithoutStandby then
		dewdrop:AddLine(
			'text', FuBarLocals.MENU_HIDE,
			'arg1', self,
			'func', "Hide",
			'level', level,
			'closeWhenClicked', true
		)
	end
	
	dewdrop:AddLine(
		'text', FuBarLocals.MENU_DISABLE,
		'arg1', self,
		'func', "DisableAddon",
		'level', level,
		'closeWhenClicked', true
	)
	
	dewdrop:AddLine(
		'level', level
	)
	
	dewdrop:AddLine(
		'text', FuBarLocals.MENU_ABOUT,
		'func', function()
			self.cmd:DisplayAddonInfo()
		end,
		'level', level
	)
end

function lib:SetIcon(path)
	assert(self.hasIcon, "Cannot set icon unless self.hasIcon is set. (" .. self:GetTitle() .. ")")
	
	if type(path) ~= "string" then
		path = "Interface\\AddOns\\" .. self.folderName .. "\\icon"
	end
	if string.sub(path, 1, 16) == "Interface\\Icons\\" then
		self.iconFrame:SetTexCoord(0.05, 0.95, 0.05, 0.95)
	else
		self.iconFrame:SetTexCoord(0, 1, 0, 1)
	end
	self.iconFrame:SetTexture(path)
end

function lib:GetIcon()
	if self.hasIcon then
		return self.iconFrame:GetTexture()
	end
end

function lib:CheckWidth(force)
	if (self.iconFrame and self.iconFrame:IsShown()) or (self.textFrame and self.textFrame:IsShown()) then
		if self.data and not self:IsIconShown() then
			self.iconFrame:SetWidth(epsilon)
		end
		local width
		if not self.hasNoText then
			self.textFrame:SetHeight(0)
			self.textFrame:SetWidth(500)
			width = self.textFrame:GetStringWidth() + 1
			self.textFrame:SetWidth(width)
			self.textFrame:SetHeight(self.textFrame:GetHeight())
		end
		if self.hasNoText or not self.textFrame:IsShown() then
			self.frame:SetWidth(self.iconFrame:GetWidth())
			if self.panel and self.panel:GetPluginSide(self) == "CENTER" then
				self.panel:UpdateCenteredPosition()
			end
		elseif force or not self.textWidth or self.textWidth < width or self.textWidth - 8 > width then
			self.textWidth = width
			self.textFrame:SetWidth(width)
			if self.iconFrame and self.iconFrame:IsShown() then
				self.frame:SetWidth(width + self.iconFrame:GetWidth())
			else
				self.frame:SetWidth(width)
			end
			if self.panel and self.panel:GetPluginSide(self) == "CENTER" then
				self.panel:UpdateCenteredPosition()
			end
		end
	end
end

function lib:SetText(text)
	assert(self.textFrame, "Cannot set text without a self.textFrame (" .. self:GetTitle() .. ")")
	assert(not self.hasNoText, "Cannot set text if self.hasNoText has been set. (" .. self:GetTitle() .. ")")
	assert(text, "You must provide a value for text, even if it is blank. (" .. self:GetTitle() .. ")")
	if text == "" then
		if self.hasIcon then
			self:ShowIcon()
		else
			text = self:GetTitle()
		end
	end
	if not self:IsTextColored() then
		text = string.gsub(text, "|c%x%x%x%x%x%x%x%x(.-)|r", "%1")
	end
	self.textFrame:SetText(text)
	self:CheckWidth()
end

function lib:GetText()
	assert(self.textFrame, "Cannot get text without a self.textFrame (" .. self:GetTitle() .. ")")
	if not self.hasNoText then
		return self.textFrame:GetText() or ""
	end
end

function lib:IsIconShown()
	if not self.hasIcon then
		return false
	elseif self.hasNoText then
		return true
	elseif not self.data then
		return true
	elseif not self.data.showIcon then
		return true
	else
		return (self.data.showIcon == 1 or self.data.showIcon == true) and true or false
	end
end

function lib:ToggleIconShown(loud)
	assert(self.iconFrame, "Cannot toggle icon without a self.iconFrame (" .. self:GetTitle() .. ")")
	assert(self.hasIcon, "Cannot show icon unless self.hasIcon is set. (" .. self:GetTitle() .. ")")
	assert(not self.hasNoText, "Cannot hide icon if self.hasNoText is set. (" .. self:GetTitle() .. ")")
	assert(self.textFrame, "Cannot hide icon if self.textFrame is not set. (" .. self:GetTitle() .. ")")
	assert(self.iconFrame, "Cannot hide icon if self.iconFrame is not set. (" .. self:GetTitle() .. ")")
	local value = not self:IsIconShown()
	self.data.showIcon = value and 1 or 0
	if loud then
		self.cmd:status(FuBarLocals.ARGUMENT_SHOWICON, value and 1 or 0, FuBarLocals.MAP_ONOFF)
	end
	if value then
		if self.textFrame:IsShown() and self.textFrame:GetText() == self:GetTitle() then
			self.textFrame:Hide()
			self.textFrame:SetText("")
		end
		self.iconFrame:Show()
		self.iconFrame:SetWidth(self.iconFrame:GetHeight())
	else
		if not self.textFrame:IsShown() or not self.textFrame:GetText() then
			self.textFrame:Show()
			self.textFrame:SetText(self:GetTitle())
		end
		self.iconFrame:Hide()
		self.iconFrame:SetWidth(epsilon)
	end
	self:CheckWidth(true)
	return value
end

function lib:ShowIcon()
	if not self:IsIconShown() then
		self:ToggleIconShown()
	end
end

function lib:HideIcon()
	if self:IsIconShown() then
		self:ToggleIconShown()
	end
end

function lib:IsTextShown()
	if not self.hasIcon then
		return false
	elseif self.hasNoText then
		return false
	elseif not self.data then
		return true
	elseif not self.data.showText then
		return true
	else
		return (self.data.showText == 1 or self.data.showText == true) and true or false
	end
end

function lib:ToggleTextShown(loud)
	assert(not self.cannotHideText, "Cannot hide text unless self.cannotHideText is unset. (" .. self:GetTitle() .. ")")
	assert(self.hasIcon, "Cannot show text unless self.hasIcon is set. (" .. self:GetTitle() .. ")")
	assert(not self.hasNoText, "Cannot hide text if self.hasNoText is set. (" .. self:GetTitle() .. ")")
	assert(self.textFrame, "Cannot hide text if self.textFrame is not set. (" .. self:GetTitle() .. ")")
	assert(self.iconFrame, "Cannot hide text if self.iconFrame is not set. (" .. self:GetTitle() .. ")")
	local value = not self:IsTextShown()
	self.data.showText = value and 1 or 0
	if loud then
		self.cmd:status(FuBarLocals.ARGUMENT_SHOWTEXT, value and 1 or 0, FuBarLocals.MAP_ONOFF)
	end
	if value then
		self.textFrame:Show()
		self:UpdateText()
	else
		self.textFrame:SetText("")
		self.textFrame:SetWidth(epsilon)
		self.textFrame:Hide()
		if not self:IsIconShown() then
			DropDownList1:Hide()
		end
		self:ShowIcon()
	end
	self:CheckWidth(true)
	return value
end

function lib:ShowText()
	if not self:IsTextShown() then
		self:ToggleTextShown()
	end
end

function lib:HideText()
	if self:IsTextShown() then
		self:ToggleTextShown()
	end
end

function lib:OpenChildFrame(child)
	assert(child, "Expected frame, received nil. (" .. self:GetTitle() .. ")")
	if FuBarPlugin.lastChildFrame then
		FuBarPlugin.lastChildFrame:Hide()
	end
	FuBarPlugin.lastChildFrame = child
	local frame = self.frame
	local width, height = GetScreenWidth(), GetScreenHeight()
	local anchor
	if frame:GetLeft() <= width/2 then
		anchor = "LEFT"
	else
		anchor = "RIGHT"
	end
	child:ClearAllPoints()
	if frame:GetTop() <= height/2 then
		child:SetPoint("BOTTOM" .. anchor, frame, "TOP" .. anchor)
		local left = child:GetLeft()
		child:ClearAllPoints()
		child:SetPoint("LEFT", UIParent, "LEFT", left, 0)
		child:SetPoint("BOTTOM", frame, "TOP")
	else
		child:SetPoint("TOP" .. anchor, frame, "BOTTOM" .. anchor)
		local left = child:GetLeft()
		child:ClearAllPoints()
		child:SetPoint("LEFT", UIParent, "LEFT", left, 0)
		child:SetPoint("TOP", frame, "BOTTOM")
	end
	child:Show()
end

function lib:HookIntoPlugin(mother, style)
	if self.hookedPlugin then
		self:debug("Cannot hook into a plugin while already hooked into another.")
		return false
	end
	if not mother or type(mother) ~= "table" or getmetatable(self).__index ~= getmetatable(mother).__index then
		self:debug("Cannot hook into given plugin.")
		return false
	end
	local showTitle = false
	if style then
		local a, b = string.find(style, "_TITLE")
		if a and b then
			showTitle = true
			style = string.sub(style, 1, a - 1) .. string.sub(style, b + 1)
		end
	end
	if style ~= "INLINE" and style ~= "SWITCH" and style ~= "SWITCHTEXT" then
		style = "INLINE"
	end
	local child = self
	local merged = false
	local current = "MOTHER"
	local mother__UpdateTooltip = mother._UpdateTooltip
	local child_UpdateTooltip = child.UpdateTooltip
	local mother_UpdateText = mother.UpdateText
	local child_UpdateText = child.UpdateText
	local child_Show = child.Show
	local mother_funcs = {}
	local mother_MenuSettings = mother.MenuSettings
	local manualStandby = false
	local motherIcon = mother:GetIcon()
	local childIcon = child:GetIcon()
	local merge, unmerge, switch
	switch = function()
		if current == "MOTHER" then
			self.data["switch." .. mother.realName] = "CHILD"
			current = "CHILD"
			getmetatable(mother).__index.SetIcon(mother, childIcon)
		else
			self.data["switch." .. mother.realName] = "MOTHER"
			current = "MOTHER"
			getmetatable(mother).__index.SetIcon(mother, motherIcon)
		end
		mother:UpdateText()
		mother:UpdateTooltip()
		dewdrop:Close()
	end
	function self:SwitchPluginHook()
		switch()
	end
	function self:IsCurrentPluginHookSwitch()
		return current ~= "MOTHER"
	end
	merge = function()
		merged = true
		self.data["merge." .. mother.realName] = 1
		if style == "INLINE" or style == "SWITCHTEXT" then
			function mother:_UpdateTooltip()
				mother__UpdateTooltip(mother)
				if showTitle and not mother:IsTooltipDetached() then
					tablet:AddCategory(
						'text', child:GetTitle(),
						'font', tablet:GetHeaderFontObject(),
						'justify', "CENTER",
						'isTitle', true
					)
				end
				child:_UpdateTooltip()
			end
		elseif style == "SWITCH" then
			function mother:_UpdateTooltip()
				if current == "MOTHER" then
					mother__UpdateTooltip(mother)
				else
					mother.tooltip:SetTitle(child:GetTitle())
					child.tooltip = mother.tooltip
					child:_UpdateTooltip()
				end
			end
		end
		function child:UpdateTooltip()
			mother:UpdateTooltip()
		end
		
		local text
		if style == "INLINE" then
			function mother:SetText(t)
				text = t
			end
			function child:SetText(t)
				if text == "" or not text then
					text = t
				else
					text = text .. " " .. t
				end
			end
		elseif style == "SWITCH" or style == "SWITCHTEXT" then
			function mother:SetText(t)
				if current == "MOTHER" then
					text = t
				end
			end
			function child:SetText(t)
				if current ~= "MOTHER" then
					text = t
				end
			end
		end
		
		if style == "INLINE" then
			function mother:UpdateText()
				text = ""
				if mother_UpdateText then
					mother_UpdateText(mother)
				end
				if child_UpdateText then
					child_UpdateText(child)
				end
				getmetatable(mother).__index.SetText(mother, text)
				text = nil
			end
		elseif style == "SWITCH" or style == "SWITCHTEXT" then
			function mother:UpdateText()
				text = ""
				if current == "MOTHER" and mother_UpdateText then
					mother_UpdateText(mother)
				end
				if current ~= "MOTHER" and child_UpdateText then
					child_UpdateText(child)
				end
				getmetatable(mother).__index.SetText(mother, text)
				text = nil
			end
		end
		child.UpdateText = mother.UpdateText
		
		for _,name in onClickTypes do
			local name = name
			mother_funcs[name] = mother[name]
			local child_func = child[name]
			if style == "INLINE" then
				if mother_funcs[name] and child_func then
					mother[name] = function(self, alpha)
						mother_funcs[name](mother, alpha)
						child_func(child, alpha)
					end
				elseif child_func then
					mother[name] = child_func
				end
			elseif style == "SWITCH" or style == "SWITCHTEXT" then
				if mother_funcs[name] and child_func then
					mother[name] = function(self, alpha)
						if current == "MOTHER" and mother_funcs[name] ~= "nil" then
							mother_funcs[name](mother, alpha)
						elseif current ~= "MOTHER" and child_func then
							child_func(child, alpha)
						end
					end
				elseif child_func then
					mother[name] = function(self, alpha)
						if current ~= "MOTHER" and child_func then
							child_func(child, alpha)
						end
					end
				elseif mother_funcs[name] then
					mother[name] = function(self, alpha)
						if current == "MOTHER" and mother_funcs[name] ~= "nil" then
							mother_funcs[name](mother, alpha)
						end
					end
				end
			end
			if not mother_funcs[name] then
				mother_funcs[name] = "nil"
			end
		end
		
		if style == "SWITCH" or style == "SWITCHTEXT" then
			function mother:SetIcon(path)
				motherIcon = path
				getmetatable(mother).__index.SetIcon(mother, path)
			end
			function child:SetIcon(path)
				childIcon = path
				getmetatable(mother).__index.SetIcon(mother, path)
			end
		end
		
		function child:Show()
			unmerge()
			
			child_Show(self)
		end
		
		if not child.hideWithoutStandby then
			manualStandby = true
			child:EnableHideWithoutStandby()
		end
		child:Hide()
		
		mother:UpdateText()
		mother:UpdateTooltip()
	end
	unmerge = function()
		merged = false
		self.data["merge." .. mother.realName] = 0
		mother._UpdateTooltip = mother__UpdateTooltip
		child.UpdateTooltip = child_UpdateTooltip
		mother.SetText = nil
		child.SetText = nil
		mother.SetIcon = nil
		child.SetIcon = nil
		mother.UpdateText = mother_UpdateText
		child.UpdateText = child_UpdateText
		for name,func in pairs(mother_funcs) do
			if func == "nil" then
				mother[name] = nil
			else
				mother[name] = func
			end
		end
		child.Show = child_Show
		if manualStandby then
			child:DisableHideWithoutStandby()
			manualStandby = false
		end
		mother:SetIcon(motherIcon)
		mother:UpdateText()
		mother:UpdateTooltip()
		child:Show()
	end
	
	function mother:MenuSettings(level, value)
		if level == 1 then
			dewdrop:AddLine(
				'text', format(FuBar.loc.PATTERN_MERGE_WITH_CHILD, child:GetTitle()),
				'func', function()
					if merged then
						unmerge()
					else
						if child.disabled then
							child:EnableAddon()
						end
						merge()
					end
				end,
				'checked', merged,
				'hasArrow', merged and (style == "SWITCH" or style == "SWITCHTEXT"),
				'value', child.realName
			)
		elseif level == 2 and value == child.realName then
			dewdrop:AddLine(
				'text', format(FuBar.loc.PATTERN_SWITCH_TO_CHILD, child:GetTitle()),
				'func', switch,
				'checked', current ~= "MOTHER"
			)
		end
		if merged then
			if (style ~= "SWITCH" or (style == "SWITCH" and current == "MOTHER")) and mother_MenuSettings then
				if level == 1 then
					dewdrop:AddLine()
				end
				mother_MenuSettings(mother, level, value)
			end
			if (style ~= "SWITCH" or (style == "SWITCH" and current == "CHILD")) and child.MenuSettings then
				if level == 1 then
					dewdrop:AddLine()
					dewdrop:AddLine(
						'text', child:GetTitle(),
						'isTitle', true
					)
				end
				child:MenuSettings(level, value)
			end
		else
			if mother_MenuSettings then
				if level == 1 then
					dewdrop:AddLine()
				end
				mother_MenuSettings(mother, level, value)
			end
		end
	end
	if not self.data["merge." .. mother.realName] then
		self.data["merge." .. mother.realName] = 1
	end
	if (style == "SWITCH" or style == "SWITCHTEXT") and not self.data["switch." .. mother.realName] then
		self.data["switch." .. mother.realName] = "MOTHER"
	end
	if self.data["merge." .. mother.realName] == 1 then
		merge()
	end
	if style == "SWITCH" or style == "SWITCHTEXT" then
		if self.data["switch." .. mother.realName] == "MOTHER" then
			self.data["switch." .. mother.realName] = "CHILD"
		else
			self.data["switch." .. mother.realName] = "MOTHER"
		end
		current = self.data["switch." .. mother.realName]
		switch()
	end
	self.hookedPlugin = mother
	return true
end

function lib:IsTooltipDetached()
	if not tablet.registry[self.frame] then
		self:UpdateTooltip()
	end
	return not tablet:IsAttached(self.frame)
end

function lib:DetachTooltip()
	tablet:Detach(self.frame)
end

function lib:ReattachTooltip()
	tablet:Attach(self.frame)
end

function lib:RegisterForLoad()
	if not self.defaults then
		self.defaults = {}
	end
	if not self.defaults.detachedTooltip then
		self.defaults.detachedTooltip = {}
	end
	
	if not self.cmd then
		local title = string.lower(self:GetTitle())
		self.cmd = AceChatCmd:new({"/" .. string.gsub(title, " ", "_"), "/" .. string.gsub(title, " ", "")}, {})
	end
	
	if not self.realName then
		if not self.name or not getglobal(self.name) then
			for i = 1, GetNumAddOns() do
				local name = GetAddOnInfo(i)
				if getglobal(name) == self then
					self.folderName = name
					self.realName = name
					break
				end
				local n = string.gsub(name, "^FuBar_", "")
				if getglobal(n) == self then
					self.folderName = name
					self.realName = n
					break
				end
			end
		end
	end
	if not self.realName then
		self.realName = string.gsub(string.gsub(self.name, " - ", "_"), " ", "")
		self.folderName = self.realName
	end
	
	self.versionNumber = FuBarUtils.VersionToNumber(self.version)
	
	if not self.frame then
		self.frame = FuBar:AcquirePluginFrame(self)
		self.textFrame = getglobal(self.frame:GetName() .. "Text")
		self.iconFrame = getglobal(self.frame:GetName() .. "Icon")
	else
		self.userDefinedFrame = true
	end
	
	self.frame.plugin = self
	self.frame:SetParent(UIParent)
	self.frame:SetPoint("RIGHT", UIParent, "LEFT", -5, 0)
	self.frame:Hide()
	
	if not self.UpdateData then
		self.UpdateData = function(self) end
	end
	if not self.UpdateText then
		if self.hasNoText then
			self.UpdateText = function(self) end
		else
			self.UpdateText = function(self)
				self:SetText(self:GetTitle())
			end
		end
	end
	if not self.UpdateTooltip then
		self.UpdateTooltip = function(self) end
	end
	
	if self.hasIcon then
		self:SetIcon(self.hasIcon)
		
		if not self.cmd.options then
			self.cmd.options = {}
		end
		
		table.insert(self.cmd.options, {
			option = FuBar.loc.ARGUMENT_SHOWICON,
			desc = FuBar.loc.TEXT_TOGGLE_WHETHER_TO_SHOW_THE_ICON,
			method = "ToggleIconShown"
		})
		
		local old_Enable = self.Enable
		function self:Enable()
			if old_Enable then
				old_Enable(self)
			end
			if self:IsIconShown() then
				self.iconFrame:Show()
			else
				self.iconFrame:Hide()
			end
			self:CheckWidth(true)
		end
		
		local old_Report = self.Report
		if old_Report then
			function self:Report()
				old_Report(self)
				ace:print(" - ", format(ACE_CMD_REPORT_LINE, FuBar.loc.ARGUMENT_SHOWICON, FuBar.loc.MAP_ONOFF[self:IsIconShown() and 1 or 0]))
			end
		else
			function self:Report()
				local report = compost:Acquire(
					compost:AcquireHash(
						'text', FuBar.loc.ARGUMENT_SHOWICON,
						'val', self:IsIconShown() and 1 or 0,
						'map', FuBar.loc.MAP_ONOFF
					)
				)
				self.cmd:report(report)
				compost:Reclaim(report, 1)
			end
		end
	end
	
	if not self.textFrame or not self.hasIcon then
		self.cannotHideText = true
	end
	if not self.cannotHideText then
		table.insert(self.cmd.options, {
			option = FuBar.loc.ARGUMENT_SHOWTEXT,
			desc = FuBar.loc.TEXT_TOGGLE_WHETHER_TO_SHOW_TEXT,
			method = "ToggleTextShown"
		})
		
		local old_Enable = self.Enable
		function self:Enable()
			if old_Enable then
				old_Enable(self)
			end
			if self:IsTextShown() then
				self.textFrame:Show()
			else
				self.textFrame:Hide()
			end
			self:CheckWidth(true)
		end
		
		local old_Report = self.Report
		if old_Report then
			function self:Report()
				old_Report(self)
				ace:print(" - ", format(ACE_CMD_REPORT_LINE, FuBar.loc.ARGUMENT_SHOWTEXT, FuBar.loc.MAP_ONOFF[self:IsTextShown() and 1 or 0]))
			end
		else
			function self:Report()
				local report = compost:Acquire(
					compost:AcquireHash(
						'text', FuBar.loc.ARGUMENT_SHOWTEXT,
						'val', self:IsTextShown() and 1 or 0,
						'map', FuBar.loc.MAP_ONOFF
					)
				)
				self.cmd:report(report)
				compost:Reclaim(report, 1)
			end
		end
	end
	
	if self.hideWithoutStandby then
		self:EnableHideWithoutStandby()
	end
	
	local old_Enable = self.Enable
	function self:Enable()
		if old_Enable then
			old_Enable(self)
		end
		if not self.hideWithoutStandby or not self.data.hidden then
			self:Show()
		end
		self:RegisterEvent("ACE_PROFILE_LOADED", "_OnAceProfileLoaded")
		
		if not self.overrideTooltip and not self.cannotDetachTooltip and self.data.detachedTooltip and self.data.detachedTooltip.detached then
			self:DetachTooltip()
		end
	end
	
	if self.db then
		local old_Initialize = self.Initialize
		function self:Initialize()
			if self.queuedDebugs then
				for _,msg in ipairs(self.queuedDebugs) do
					self:debug(msg)
				end
				compost:Reclaim(self.queuedDebugs)
				self.queuedDebugs = nil
			end
			local t = self.db._table
			if not t.profiles then
				t.profiles = {}
			end
			if not t.profiles[self.profilePath[2]] then
				t.profiles[self.profilePath[2]] = {}
			end
			
			if not t.realms then
				t.realms = {}
			end
			if not t.realms[GetRealmName() .. UnitFactionGroup("player")] then
				t.realms[GetRealmName() .. UnitFactionGroup("player")] = {}
			end
			if type(self.chardb) ~= "string" then
				if not t.chars then
					t.chars = {}
				end
				if not t.chars[GetRealmName() .. UnitName("player")] then
					t.chars[GetRealmName() .. UnitName("player")] = {}
				end
				self.charData = t.chars[GetRealmName() .. UnitName("player")]
			else
				local c = getglobal(self.chardb)
				if type(c) ~= "table" then
					setglobal(self.chardb, {})
					c = getglobal(self.chardb)
				end
				self.charData = c
			end
			if not t.classes then
				t.classes = {}
			end
			if not t.classes[UnitClass("player")] then
				t.classes[UnitClass("player")] = {}
			end
			if not t.full then
				t.full = {}
			end
			self.data = t.profiles[self.profilePath[2]]
			inheritDefaults(self.data, self.defaults)
			self.realmData = t.realms[GetRealmName() .. UnitFactionGroup("player")]
			inheritDefaults(self.realmData, self.realmDefaults)
			inheritDefaults(self.charData, self.charDefaults)
			self.classData = t.classes[UnitClass("player")]
			inheritDefaults(self.classData, self.classDefaults)
			self.fullData = t.full
			inheritDefaults(self.fullData, self.fullDefaults)
			
			self.data.version = FuBarUtils.VersionToNumber(self.data.version)
			self.charData.version = FuBarUtils.VersionToNumber(self.charData.version)
			self.classData.version = FuBarUtils.VersionToNumber(self.classData.version)
			self.realmData.version = FuBarUtils.VersionToNumber(self.realmData.version)
			self.fullData.version = FuBarUtils.VersionToNumber(self.fullData.version)
			if self.data.version >= 500 then
				self.data.version = self.data.version / 100
				self.charData.version = self.charData.version / 100
				self.classData.version = self.classData.version / 100
				self.realmData.version = self.realmData.version / 100
				self.fullData.version = self.fullData.version / 100
			end
			
			if old_Initialize then
				old_Initialize(self)
			end
			
			self.data.version = FuBarUtils.VersionToNumber(self.version)
			self.charData.version = FuBarUtils.VersionToNumber(self.version)
			self.classData.version = FuBarUtils.VersionToNumber(self.version)
			self.realmData.version = FuBarUtils.VersionToNumber(self.version)
			self.fullData.version = FuBarUtils.VersionToNumber(self.version)
		end
	else
		local old_Initialize = self.Initialize
		function self:Initialize()
			if self.queuedDebugs then
				for _,msg in ipairs(self.queuedDebugs) do
					self:debug(msg)
				end
				self.queuedDebugs = nil
			end
			
			if not FuBarDB.pluginDB then
				FuBarDB.pluginDB = {}
			end
			
			if not FuBarDB.pluginDB[self.realName] then
				FuBarDB.pluginDB[self.realName] = {}
			end
			
			local t = FuBarDB.pluginDB[self.realName]
			
			if not t.profiles then
				t.profiles = {}
			end
			if not t.profiles[FuBar.profilePath[2]] then
				t.profiles[FuBar.profilePath[2]] = {}
			end
			if not t.realms then
				t.realms = {}
			end
			if not t.realms[GetRealmName() .. UnitFactionGroup("player")] then
				t.realms[GetRealmName() .. UnitFactionGroup("player")] = {}
			end
			if type(self.chardb) ~= "string" then
				if not t.chars then
					t.chars = {}
				end
				if not t.chars[GetRealmName() .. UnitName("player")] then
					t.chars[GetRealmName() .. UnitName("player")] = {}
				end
				self.charData = t.chars[GetRealmName() .. UnitName("player")]
			else
				local c = getglobal(self.chardb)
				if type(c) ~= "table" then
					setglobal(self.chardb, {})
					c = getglobal(self.chardb)
				end
				self.charData = c
			end
			if not t.classes then
				t.classes = {}
			end
			if not t.classes[UnitClass("player")] then
				t.classes[UnitClass("player")] = {}
			end
			if not t.full then
				t.full = {}
			end
			self.data = t.profiles[FuBar.profilePath[2]]
			inheritDefaults(self.data, self.defaults)
			self.realmData = t.realms[GetRealmName() .. UnitFactionGroup("player")]
			inheritDefaults(self.realmData, self.realmDefaults)
			inheritDefaults(self.charData, self.charDefaults)
			self.classData = t.classes[UnitClass("player")]
			inheritDefaults(self.classData, self.classDefaults)
			self.fullData = t.full
			inheritDefaults(self.fullData, self.fullDefaults)
			
			if old_Initialize then
				old_Initialize(self)
			end
			
			self.data.version = FuBarUtils.VersionToNumber(self.version)
			self.charData.version = FuBarUtils.VersionToNumber(self.version)
			self.classData.version = FuBarUtils.VersionToNumber(self.version)
			self.realmData.version = FuBarUtils.VersionToNumber(self.version)
			self.fullData.version = FuBarUtils.VersionToNumber(self.version)
			
			if self.data.disabled then
				self:DisableAddon()
			end
		end
	end
	
	local old_Disable = self.Disable
	function self:Disable()
		if old_Disable then
			old_Disable(self)
		end
		self:Hide()
	end
	
	if not self.overrideTooltip then
		self._UpdateTooltip = self.UpdateTooltip
		local first = true
		function self:UpdateTooltip()
			if first then
				first = false
				tablet:Register(self.frame,
					'children', function()
						tablet:SetTitle(self:GetTitle())
						self:_UpdateTooltip()
					end,
					'clickable', self.clickableTooltip,
					'data', FuBar.data.tooltip,
					'detachedData', self.data.detachedTooltip,
					'point', function(frame)
						if frame:GetTop() > GetScreenHeight() / 2 then
							local x = frame:GetCenter()
							if x < GetScreenWidth() / 3 then
								return "TOPLEFT", "BOTTOMLEFT"
							elseif x < GetScreenWidth() * 2 / 3 then
								return "TOP", "BOTTOM"
							else
								return "TOPRIGHT", "BOTTOMRIGHT"
							end
						else
							local x = frame:GetCenter()
							if x < GetScreenWidth() / 3 then
								return "BOTTOMLEFT", "TOPLEFT"
							elseif x < GetScreenWidth() * 2 / 3 then
								return "BOTTOM", "TOP"
							else
								return "BOTTOMRIGHT", "TOPRIGHT"
							end
						end
					end,
					'menu', function(level, value, valueN_1, valueN_2, valueN_3, valueN_4)
						self:MenuSettings(level, value, true, valueN_1, valueN_2, valueN_3, valueN_4)
					end
				)
			end
			tablet:Refresh(self.frame)
		end
	end
	
	local self_Initialize = self.Initialize
	function self:Initialize()
		self_Initialize(self)
		
		local profilePlugin = FuBar:GetPluginProfiling()
		
		if self == profilePlugin then
			local start = GetTime()
			local tree = {}
			local treeMemories = {}
			local treeTimes = {}
			local memories = {}
			local times = {}
			local aname = self.name
			for name,value in pairs(self) do
				if type(value) == "function" and name ~= "Initialize" then
					local oldFunction = value
					local name = name
					local rName = name
					if name == "UpdateTooltip" then
						name = "tablet:Refresh"
					elseif name == "_UpdateTooltip" then
						name = "UpdateTooltip"
					end
					if strsub(name, 1, 1) ~= "_" then
						memories[name] = 0
						times[name] = 0
						self[rName] = function(self, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20)
							local pos = table.getn(tree)
							table.insert(tree, name)
							table.insert(treeMemories, 0)
							table.insert(treeTimes, 0)
							local t, mem = GetTime(), gcinfo() 
							local r1, r2, r3, r4, r5, r6, r7, r8 = oldFunction(self, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a20)
							mem, t = gcinfo() - mem, GetTime() - t
							if pos > 0 then
								treeMemories[pos] = treeMemories[pos] + mem
								treeTimes[pos] = treeTimes[pos] + t
							end
							local otherMem = table.remove(treeMemories)
							if mem - otherMem > 0 then
								memories[name] = memories[name] + mem - otherMem
							end
							times[name] = times[name] + t - table.remove(treeTimes)
							table.remove(tree)
							return r1, r2, r3, r4, r5, r6, r7, r8
						end
					end
				end
			end
			
			if not self.cmd.options then
				self.cmd.options = {}
			end
			
			table.insert(self.cmd.options, {
				option = FuBar.loc.ARGUMENT_PROFILE,
				desc = FuBar.loc.TEXT_LIST_PROFILING_INFO,
				method = "_ListProfileInfo"
			})
			
			function self:_ListProfileInfo()
				local duration = GetTime() - start
				for name in pairs(memories) do
					local mem = memories[name]
					local time = times[name]
					DEFAULT_CHAT_FRAME:AddMessage(string.format("%.3fs / %.3fs (%.3f%%) || %s (%d KiB)", times[name], duration, times[name] / duration * 100, name, memories[name]))
				end
			end
			
			function self:GetProfileInfo()
				return GetTime() - start, times, memories
			end
		end
	end
	
	if self.defaultPosition ~= "LEFT" and self.defaultPosition ~= "CENTER" and self.defaultPosition ~= "RIGHT" and self.defaultPosition ~= "MINIMAP" then
		self.defaultPosition = "LEFT"
	end
	if type(self.defaultMinimapPosition) ~= "number" then
		self.defaultMinimapPosition = math.random(1, 360)
	end
	AceAddon.RegisterForLoad(self)
	FuBar:RegisterPlugin(self)
end

function lib:GetLibraryVersion()
	return MAJOR_VERSION, MINOR_VERSION
end

function lib:LibActivate(stub, oldLib, oldList)
end

function lib:LibDeactivate()
end

FuBarPlugin:Register(lib)
lib = nil

if not FuBarPlugin.new then
	function FuBarPlugin:new(o)
		assert(type(o) == "table", "You must provide a table to register")
		local name
		if o.name then
			name = o.name .. ": "
		else
			name = ""
		end
		assert(o.fuCompatible, name .. "You must provide a fuCompatible field.")
		local compatNumber = FuBarUtils.VersionToNumber(o.fuCompatible)
		if compatNumber >= 500 then
			compatNumber = compatNumber / 100
		end
		if compatNumber < 101 then
			return FuBarPlugin:GetInstance("1.0"):new(o)
		elseif compatNumber < 103 then
			return FuBarPlugin:GetInstance("1.2"):new(o)
		else
			return FuBarPlugin:GetInstance("1.4"):new(o)
		end
	end
end
