local compost = CompostLib:GetInstance('compost-1')
local babble = BabbleLib:GetInstance('1.0')

FuBarUtils = nil

FUBAR_REVISION = tonumber(string.sub("$Revision: 1637 $", 12, -3))

local COLOR_HEX_RED       = "ff0000"
local COLOR_HEX_ORANGE    = "ff7f00"
local COLOR_HEX_YELLOW    = "ffff00"
local COLOR_HEX_GREEN     = "00ff00"
local COLOR_HEX_WHITE     = "ffffff"
local COLOR_HEX_COPPER    = "eda55f"
local COLOR_HEX_SILVER    = "c7c7cf"
local COLOR_HEX_GOLD      = "ffd700"

local Colorize = function(hexColor, text)
	return "|cff" .. tostring(hexColor or 'ffffff') .. tostring(text) .. "|r"
end
local function Red(text) return Colorize(COLOR_HEX_RED, text) end
local function Orange(text) return Colorize(COLOR_HEX_ORANGE, text) end
local function Yellow(text) return Colorize(COLOR_HEX_YELLOW, text) end
local function Green(text) return Colorize(COLOR_HEX_GREEN, text) end
local function White(text) return Colorize(COLOR_HEX_WHITE, text) end
local function Copper(text) return Colorize(COLOR_HEX_COPPER, text) end
local function Silver(text) return Colorize(COLOR_HEX_SILVER, text) end
local function Gold(text) return Colorize(COLOR_HEX_GOLD, text) end

local nan = 0/0
local inf = 1/0
local ninf = -1/0

local function GetThresholdColor(quality, worst, worse, normal, better, best)
	if quality == nan or quality == inf or quality == ninf or tonumber(quality) == nil then
		return 1, 1, 1
	end
	if worst == nil then
		worst = 0
		worse = 0.25
		normal = 0.5
		better = 0.75
		best = 1
	end
	
	if worst < best then
		if (worse == better and quality == worse) or (worst == best and quality == worst) then
			return 1, 1, 0
		elseif quality <= worst then
			return 1, 0, 0
		elseif quality <= worse then
			return 1, 0.5 * (quality - worst) / (worse - worst), 0
		elseif quality <= normal then
			return 1, 0.5 + 0.5 * (quality - worse) / (normal - worse), 0
		elseif quality <= better then
			return 1 - 0.5 * (quality - normal) / (better - normal), 1, 0
		elseif quality <= best then
			return 0.5 - 0.5 * (quality - better) / (best - better), 1, 0
		else
			return 0, 1, 0
		end
	else
		if (worse == better and quality == worse) or (worst == best and quality == worst) then
			return 1, 1, 0
		elseif quality >= worst then
			return 1, 0, 0
		elseif quality >= worse then
			return 1, 0.5 - 0.5 * (quality - worse) / (worst - worse), 0
		elseif quality >= normal then
			return 1, 1 - 0.5 * (quality - normal) / (worse - normal), 0
		elseif quality >= better then
			return 0.5 + 0.5 * (quality - better) / (normal - better), 1, 0
		elseif quality >= best then
			return 0.5 * (quality - best) / (better - best), 1, 0
		else
			return 0, 1, 0
		end
	end
end

local function GetThresholdHexColor(quality, worst, worse, normal, better, best)
	local r, g, b = GetThresholdColor(quality, worst, worse, normal, better, best)
	return format("%02x%02x%02x", r*255, g*255, b*255)
end

local function GetThresholdColorTrivial(quality, worst, worse, normal, better, best)
	if quality == nan or quality == inf or quality == ninf or tonumber(quality) == nil then
		return 1, 1, 1
	end
	if worst == nil then
		worst = 0
		worse = 0.25
		normal = 0.5
		better = 0.75
		best = 1
	end
	
	if worst < best then
		if worse == better and normal == worse then
			return 1, 1, 0
		elseif quality <= worst then
			return 1, 0, 0
		elseif quality <= worse then
			return 1, 0.5 * (quality - worst) / (worse - worst), 0
		elseif quality <= normal then
			return 1, 0.5 + 0.5 * (quality - worse) / (normal - worse), 0
		elseif quality <= better then
			return 1 - (quality - normal) / (better - normal), 1, 0
		elseif quality <= best then
			local x = 0.5 * (quality - better) / (best - better)
			return x, 1 - x, x
		else
			return 0.5, 0.5, 0.5
		end
	else
		if worse == better and normal == worse then
			return 1, 1, 0
		elseif quality >= worst then
			return 1, 0, 0
		elseif quality >= worse then
			return 1, 0.5 - 0.5 * (quality - worse) / (worst - worse), 0
		elseif quality >= normal then
			return 1, 1 - 0.5 * (quality - normal) / (worse - normal), 0
		elseif quality >= better then
			return (quality - better) / (normal - better), 1, 0
		elseif quality >= best then
			local x = 0.5 * (quality - best) / (better - best)
			return 0.5 - x, 0.5 + x, 0.5 - x
		else
			return 0.5, 0.5, 0.5
		end
	end
end

local function GetThresholdHexColorTrivial(quality, worst, worse, normal, better, best)
	local r, g, b = GetThresholdColorTrivial(quality, worst, worse, normal, better, best)
	return format("%02x%02x%02x", r*255, g*255, b*255)
end

local COPPER_ABBR = strlower(strsub(COPPER, 1, 1))
local SILVER_ABBR = strlower(strsub(SILVER, 1, 1))
local GOLD_ABBR = strlower(strsub(GOLD, 1, 1))

local function FormatMoneyExtended(value, colorize, textColor)
	local gold = value / 10000
	local silver = mod(value / 100, 100)
	local copper = mod(value, 100)
	
	local color = COLOR_HEX_WHITE
	if textColor then
		if value > 0 then
			color = COLOR_HEX_GREEN
		elseif value < 0 then
			color = COLOR_HEX_RED
		end
	end
	if colorize then
		if value >= 10000 or value <= -10000 then
			return format("|cff%s%d|r|cff%s %s|r |cff%s%d|r|cff%s %s|r |cff%s%d|r|cff%s %s|r", color, gold, COLOR_HEX_GOLD, GOLD, color, silver, COLOR_HEX_SILVER, SILVER, color, copper, COLOR_HEX_COPPER, COPPER)
		elseif value >= 100 or value <= -100 then
			return format("|cff%s%d|r|cff%s %s|r |cff%s%d|r|cff%s %s|r", color, silver, COLOR_HEX_SILVER, SILVER, color, copper, COLOR_HEX_COPPER, COPPER)
		else
			return format("|cff%s%d|r|cff%s %s|r", color, copper, COLOR_HEX_COPPER, COPPER)
		end
	else
		if value >= 10000 or value <= -10000 then
			return format("%d %s %d %s %d %s", gold, GOLD, silver, SILVER, copper, COPPER)
		elseif value >= 100 or value <= -100 then
			return format("%d %s %d %s", silver, SILVER, copper, COPPER)
		else
			return format("%d %s", copper, COPPER)
		end
	end
end

local function FormatMoneyFull(value, colorize, textColor)
	local gold = value / 10000
	local silver = mod(value / 100, 100)
	local copper = mod(value, 100)
	
	local color = COLOR_HEX_WHITE
	if textColor then
		if value > 0 then
			color = COLOR_HEX_GREEN
		elseif value < 0 then
			color = COLOR_HEX_RED
		end
	end
	if colorize then
		if value >= 10000 or value <= -10000 then
			return format("|cff%s%d|r|cff%s%s|r |cff%s%d|r|cff%s%s|r |cff%s%d|r|cff%s%s|r", color, gold, COLOR_HEX_GOLD, GOLD_ABBR, color, silver, COLOR_HEX_SILVER, SILVER_ABBR, color, copper, COLOR_HEX_COPPER, COPPER_ABBR)
		elseif value >= 100 or value <= -100 then
			return format("|cff%s%d|r|cff%s%s|r |cff%s%d|r|cff%s%s|r", color, silver, COLOR_HEX_SILVER, SILVER_ABBR, color, copper, COLOR_HEX_COPPER, COPPER_ABBR)
		else
			return format("|cff%s%d|r|cff%s%s|r", color, copper, COLOR_HEX_COPPER, COPPER_ABBR)
		end
	else
		if value >= 10000 or value <= -10000 then
			return format("%d%s %d%s %d%s", gold, GOLD_ABBR, silver, SILVER_ABBR, copper, COPPER_ABBR)
		elseif value >= 100 or value <= -100 then
			return format("%d%s %d%s", silver, SILVER_ABBR, copper, COPPER_ABBR)
		else
			return format("%d%s", copper, COPPER_ABBR)
		end
	end
end

local function FormatMoneyShort(copper, colorize, textColor)
	local color = COLOR_HEX_WHITE
	if textColor then
		if copper > 0 then
			color = COLOR_HEX_GREEN
		elseif copper < 0 then
			color = COLOR_HEX_RED
		end
	end
	if colorize then
		if copper >= 10000 or copper <= -10000 then
			return format("|cff%s%.1f|r|cff%s%s|r", color, copper / 10000, COLOR_HEX_GOLD, GOLD_ABBR)
		elseif copper >= 100 or copper <= -100 then
			return format("|cff%s%.1f|r|cff%s%s|r", color, copper / 100, COLOR_HEX_SILVER, SILVER_ABBR)
		else
			return format("|cff%s%d|r|cff%s%s|r", color, copper, COLOR_HEX_COPPER, COPPER_ABBR)
		end
	else
		if copper >= 10000 or copper <= -10000 then
			return format("%.1f%s", copper / 10000, GOLD_ABBR)
		elseif copper >= 100 or copper <= -100 then
			return format("%.1f%s", copper / 100, SILVER_ABBR)
		else
			return format("%.0f%s", copper, COPPER_ABBR)
		end
	end
end

local function FormatMoneyCondensed(value, colorize, textColor)
	local negl = value < 0 and "(-" or ""
	local negr = value < 0 and ")" or ""
	if value < 0 then
		if colorize and textColor then
			negl = "|cffff0000-(|r"
			negr = "|cffff0000)|r"
		else
			negl = "-("
			negr = ")"
		end
	else
		negl = ""
		negr = ""
	end
	local gold = floor(math.abs(value) / 10000)
	local silver = mod(floor(math.abs(value) / 100), 100)
	local copper = mod(floor(math.abs(value)), 100)
	if colorize then
		if gold ~= 0 then
			return negl .. format("|cff%s%d|r.|cff%s%02d|r.|cff%s%02d|r", COLOR_HEX_GOLD, gold, COLOR_HEX_SILVER, silver, COLOR_HEX_COPPER, copper) .. negr
		elseif silver ~= 0 then
			return negl .. format("|cff%s%d|r.|cff%s%02d|r", COLOR_HEX_SILVER, silver, COLOR_HEX_COPPER, copper) .. negr
		else
			return negl .. format("|cff%s%d|r", COLOR_HEX_COPPER, copper) .. negr
		end
	else
		if gold ~= 0 then
			return negl .. format("%d.%02d.%02d", gold, silver, copper) .. negr
		elseif silver ~= 0 then
			return negl .. format("%d.%02d", silver, copper) .. negr
		else
			return negl .. format("%dc", copper) .. negr
		end
	end
end

local function FormatDurationExtended(duration, colorize, hideSeconds)
	local negative = ""
	if duration < 0 then
		negative = "-"
		duration = -duration
	end
	local days = floor(duration / 86400)
	local hours = mod(floor(duration / 3600), 24)
	local mins = mod(floor(duration / 60), 60)
	local secs = mod(floor(duration), 60)
	if not colorize then
		if duration == nil or duration > 86400*365 then
			return "Undetermined"
		end
		local t = compost:Acquire()
		if days > 1 then
			table.insert(t, format("%d %s", days, DAYS_ABBR_P1))
		elseif days == 1 then
			table.insert(t, format("%d %s", days, DAYS_ABBR))
		end
		if hours > 1 then
			table.insert(t, format("%d %s", hours, HOURS_ABBR_P1))
		elseif hours == 1 then
			table.insert(t, format("%d %s", hours, HOURS_ABBR))
		end
		if mins > 1 then
			table.insert(t, format("%d %s", mins, MINUTES_ABBR_P1))
		elseif mins == 1 then
			table.insert(t, format("%d %s", mins, MINUTES_ABBR))
		end
		if not hideSeconds then
			if secs > 1 then
				table.insert(t, format("%d %s", secs, SECONDS_ABBR_P1))
			elseif secs == 1 then
				table.insert(t, format("%d %s", secs, SECONDS_ABBR))
			end
		end
		if table.getn(t) == 0 then
			compost:Reclaim(t)
			if not hideSeconds then
				return "0 " .. SECONDS_ABBR_P1
			else
				return "0 " .. MINUTES_ABBR_P1
			end
		else
			local s = negative .. table.concat(t, " ")
			compost:Reclaim(t)
			return s
		end
	else
		if duration == nil or duration > 86400*365 then
			return "|cffffffffUndetermined|r"
		end
		local t = compost:Acquire()
		if days > 1 then
			table.insert(t, format("|cffffffff%d|r %s", days, DAYS_ABBR_P1))
		elseif days == 1 then
			table.insert(t, format("|cffffffff%d|r %s", days, DAYS_ABBR))
		end
		if hours > 1 then
			table.insert(t, format("|cffffffff%d|r %s", hours, HOURS_ABBR_P1))
		elseif hours == 1 then
			table.insert(t, format("|cffffffff%d|r %s", hours, HOURS_ABBR))
		end
		if mins > 1 then
			table.insert(t, format("|cffffffff%d|r %s", mins, MINUTES_ABBR_P1))
		elseif mins == 1 then
			table.insert(t, format("|cffffffff%d|r %s", mins, MINUTES_ABBR))
		end
		if not hideSeconds then
			if secs > 1 then
				table.insert(t, format("|cffffffff%d|r %s", secs, SECONDS_ABBR_P1))
			elseif secs == 1 then
				table.insert(t, format("|cffffffff%d|r %s", secs, SECONDS_ABBR))
			end
		end
		if table.getn(t) == 0 then
			compost:Reclaim(t)
			if not hideSeconds then
				return "|cffffffff0|r " .. SECONDS_ABBR_P1
			else
				return "|cffffffff0|r " .. MINUTES_ABBR_P1
			end
		elseif negative == "-" then
			local s = "|cffffffff-|r" .. table.concat(t, " ")
			compost:Reclaim(t)
			return s
		else
			local s = table.concat(t, " ")
			compost:Reclaim(t)
			return s
		end
	end
end

local function FormatDurationFull(duration, colorize, hideSeconds)
	local negative = ""
	if duration < 0 then
		negative = "-"
		duration = -duration
	end
	if not colorize then
		if not hideSeconds then
			if duration == nil or duration > 86400*365 then
				return "Undetermined"
			elseif duration >= 86400 then
				return format("%s%d%s %02d%s %02d%s %02d%s", negative, duration/86400, DAY_ONELETTER_ABBR, mod(duration/3600, 24), HOUR_ONELETTER_ABBR, mod(duration/60, 60), MINUTE_ONELETTER_ABBR, mod(duration, 60), SECOND_ONELETTER_ABBR)
			elseif duration >= 3600 then
				return format("%s%d%s %02d%s %02d%s", negative, duration/3600, HOUR_ONELETTER_ABBR, mod(duration/60, 60), MINUTE_ONELETTER_ABBR, mod(duration, 60), SECOND_ONELETTER_ABBR)
			elseif duration >= 120 then
				return format("%s%d%s %02d%s", negative, duration/60, MINUTE_ONELETTER_ABBR, mod(duration, 60), SECOND_ONELETTER_ABBR)
			else
				return format("%s%d%s", negative, duration, SECOND_ONELETTER_ABBR)
			end
		else
			if duration == nil or duration > 86400*365 then
				return "Undetermined"
			elseif duration >= 86400 then
				return format("%s%d%s %02d%s %02d%s", negative, duration/86400, DAY_ONELETTER_ABBR, mod(duration/3600, 24), HOUR_ONELETTER_ABBR, mod(duration/60, 60), MINUTE_ONELETTER_ABBR)
			elseif duration >= 3600 then
				return format("%s%d%s %02d%s", negative, duration/3600, HOUR_ONELETTER_ABBR, mod(duration/60, 60), MINUTE_ONELETTER_ABBR)
			else
				return format("%s%d%s", negative, duration/60, MINUTE_ONELETTER_ABBR)
			end
		end
	else
		if not hideSeconds then
			if duration == nil or duration > 86400*365 then
				return "|cffffffffUndetermined|r"
			elseif duration >= 86400 then
				return format("|cffffffff%s%d|r%s |cffffffff%02d|r%s |cffffffff%02d|r%s |cffffffff%02d|r%s", negative, duration/86400, DAY_ONELETTER_ABBR, mod(duration/3600, 24), HOUR_ONELETTER_ABBR, mod(duration/60, 60), MINUTE_ONELETTER_ABBR, mod(duration, 60), SECOND_ONELETTER_ABBR)
			elseif duration >= 3600 then
				return format("|cffffffff%s%d|r%s |cffffffff%02d|r%s |cffffffff%02d|r%s", negative, duration/3600, HOUR_ONELETTER_ABBR, mod(duration/60, 60), MINUTE_ONELETTER_ABBR, mod(duration, 60), SECOND_ONELETTER_ABBR)
			elseif duration >= 120 then
				return format("|cffffffff%s%d|r%s |cffffffff%02d|r%s", negative, duration/60, MINUTE_ONELETTER_ABBR, mod(duration, 60), SECOND_ONELETTER_ABBR)
			else
				return format("|cffffffff%s%d|r%s", negative, duration, SECOND_ONELETTER_ABBR)
			end
		else
			if duration == nil or duration > 86400*365 then
				return "|cffffffffUndetermined|r"
			elseif duration >= 86400 then
				return format("|cffffffff%s%d|r%s |cffffffff%02d|r%s |cffffffff%02d|r%s", negative, duration/86400, DAY_ONELETTER_ABBR, mod(duration/3600, 24), HOUR_ONELETTER_ABBR, mod(duration/60, 60), MINUTE_ONELETTER_ABBR)
			elseif duration >= 3600 then
				return format("|cffffffff%s%d|r%s |cffffffff%02d|r%s", negative, duration/3600, HOUR_ONELETTER_ABBR, mod(duration/60, 60), MINUTE_ONELETTER_ABBR)
			else
				return format("|cffffffff%s%d|r%s", negative, duration/60, MINUTE_ONELETTER_ABBR)
			end
		end
	end
end

local function FormatDurationShort(duration, colorize, hideSeconds)
	local negative = ""
	if duration < 0 then
		negative = "-"
		duration = -duration
	end
	if not colorize then
		if duration == nil or duration >= 86400*365 then
			return "***"
		elseif duration >= 172800 then
			return format("%s%.1f %s", negative, duration/86400, DAYS_ABBR_P1)
		elseif duration >= 7200 then
			return format("%s%.1f %s", negative, duration/3600, HOURS_ABBR_P1)
		elseif duration >= 120 or not hideSeconds then
			return format("%s%.1f %s", negative, duration/60, MINUTES_ABBR_P1)
		else
			return format("%s%.0f %s", negative, duration, SECONDS_ABBR_P1)
		end
	else
		if duration == nil or duration >= 86400*365 then
			return "|cffffffff***|r"
		elseif duration >= 172800 then
			return format("|cffffffff%s%.1f|r %s", negative, duration/86400, DAYS_ABBR_P1)
		elseif duration >= 7200 then
			return format("|cffffffff%s%.1f|r %s", negative, duration/3600, HOURS_ABBR_P1)
		elseif duration >= 120 or not hideSeconds then
			return format("|cffffffff%s%.1f|r %s", negative, duration/60, MINUTES_ABBR_P1)
		else
			return format("|cffffffff%s%.0f|r %s", negative, duration, SECONDS_ABBR_P1)
		end
	end
end

local function FormatDurationCondensed(duration, colorize, hideSeconds)
	local negative = ""
	if duration < 0 then
		negative = "-"
		duration = -duration
	end
	if not colorize then
		if hideSeconds then
			if duration == nil or duration >= 86400*365 then
				return negative .. "**" .. DAY_ONELETTER_ABBR .. " **:**"
			elseif duration >= 86400 then
				return format("%s%d%s %d:%02d", negative, duration/86400, DAY_ONELETTER_ABBR, mod(duration/3600, 24), mod(duration/60, 60))
			else
				return format("%s%d:%02d", negative, duration/3600, mod(duration/60, 60))
			end
		else
			if duration == nil or duration >= 86400*365 then
				return negative .. "**:**:**:**"
			elseif duration >= 86400 then
				return format("%s%d%s %d:%02d:%02d", negative, duration/86400, DAY_ONELETTER_ABBR, mod(duration/3600, 24), mod(duration/60, 60), mod(duration, 60))
			elseif duration >= 3600 then
				return format("%s%d:%02d:%02d", negative, duration/3600, mod(duration/60, 60), mod(duration, 60))
			else
				return format("%s%d:%02d", negative, duration/60, mod(duration, 60))
			end
		end
	else
		if hideSeconds then
			if duration == nil or duration >= 86400*365 then
				return "|cffffffff" .. negative .. "**|r" .. DAY_ONELETTER_ABBR .. " |cffffffff**|r:|cffffffff**|r"
			elseif duration >= 86400 then
				return format("|cffffffff%s%d|r%s |cffffffff%d|r:|cffffffff%02d|r", negative, duration/86400, DAY_ONELETTER_ABBR, mod(duration/3600, 24), mod(duration/60, 60))
			else
				return format("|cffffffff%s%d|r:|cffffffff%02d|r", negative, duration/3600, mod(duration/60, 60))
			end
		else
			if duration == nil or duration >= 86400*365 then
				return "|cffffffff" .. negative .. "**|r" .. DAY_ONELETTER_ABBR .. " |cffffffff**|r:|cffffffff**|r:|cffffffff**|r"
			elseif duration >= 86400 then
				return format("|cffffffff%s%d|r%s |cffffffff%d|r:|cffffffff%02d|r:|cffffffff%02d|r", negative, duration/86400, DAY_ONELETTER_ABBR, mod(duration/3600, 24), mod(duration/60, 60), mod(duration, 60))
			elseif duration >= 3600 then
				return format("|cffffffff%s%d|r:|cffffffff%02d|r:|cffffffff%02d|r", negative, duration/3600, mod(duration/60, 60), mod(duration, 60))
			else
				return format("|cffffffff%s%d|r:|cffffffff%02d|r", negative, duration/60, mod(duration, 60))
			end
		end
	end
end

local function VersionToNumber(version)
	if version == nil then
		return 0
	elseif type(version) == "number" then
		return version
	end
	local num = 0
	local place = 1
	string.gsub(version, "(%d+)", function (w)
		num = num + tonumber(w) * math.pow(10, place * 2)
		if place > 0 then
			place = place - 1
		else
			place = place - 2
		end
	end)
	return num
end

local function GetVersionPoint(version, point)
	if version == nil then
		return 0
	elseif type(version) == "number" then
		if point == 1 then
			return floor(version / 100)
		elseif point == 2 then
			return mod(floor(version), 100)
		elseif point == 3 then
			return mod(floor(version * 10000), 10000)
		elseif point == 4 then
			return mod(floor(version * 100000000), 10000)
		elseif point < 1 then
			return 0
		elseif point > 4 then
			return mod(floor(version * math.pow(10000, point - 2)), 10000)
		end
	end
	local place = 1
	local num = 0
	string.gsub(version, "(%d+)", function (w)
		if place == point
			then num = tonumber(w)
			return
		end
		place = place + 1
	end)
	return num
end

local function GetMajorNumber(version)
	return GetVersionPoint(version, 1)
end

local function GetMinorNumber(version)
	return GetVersionPoint(version, 2)
end

local function GetRevisionNumber(version)
	return GetVersionPoint(version, 3)
end

local function GetTrivialNumber(version)
	return GetVersionPoint(version, 4)
end

FuBarUtils = {
	COLOR_HEX_RED = COLOR_HEX_RED,
	COLOR_HEX_ORANGE = COLOR_HEX_ORANGE,
	COLOR_HEX_YELLOW = COLOR_HEX_YELLOW,
	COLOR_HEX_GREEN = COLOR_HEX_GREEN,
	COLOR_HEX_WHITE = COLOR_HEX_WHITE,
	COLOR_HEX_COPPER = COLOR_HEX_COPPER,
	COLOR_HEX_SILVER = COLOR_HEX_SILVER,
	COLOR_HEX_GOLD = COLOR_HEX_GOLD,
	
	Colorize = Colorize,
	Red = Red,
	Orange = Orange,
	Yellow = Yellow,
	Green = Green,
	White = White,
	GetThresholdHexColor = GetThresholdHexColor,
	GetThresholdColor = GetThresholdColor,
	GetThresholdHexColorTrivial = GetThresholdHexColorTrivial,
	GetThresholdColorTrivial = GetThresholdColorTrivial,
	FormatMoneyExtended = FormatMoneyExtended,
	FormatMoneyFull = FormatMoneyFull,
	FormatMoneyShort = FormatMoneyShort,
	FormatMoneyCondensed = FormatMoneyCondensed,
	FormatDurationExtended = FormatDurationExtended,
	FormatDurationFull = FormatDurationFull,
	FormatDurationShort = FormatDurationShort,
	FormatDurationCondensed = FormatDurationCondensed,
	VersionToNumber = VersionToNumber,
	GetVersionPoint = GetVersionPoint,
	GetMajorNumber = GetMajorNumber,
	GetMinorNumber = GetMinorNumber,
	GetRevisionNumber = GetRevisionNumber,
	GetTrivialNumber = GetTrivialNumber,
}

local old_UIDropDownMenu_AddButton = UIDropDownMenu_AddButton
local spacerDropDown = nil
function FuBarUtils.AddSpacerDropDownMenu(level)
	if spacerDropDown ~= (level or 1) then
		spacerDropDown = level or 1
		local info = compost:AcquireHash(
			'disabled', true
		)
		old_UIDropDownMenu_AddButton(info, level or 1)
		compost:Reclaim(info)
	end
end

UIDropDownMenu_AddButton = function(info, level)
	spacerDropDown = nil
	old_UIDropDownMenu_AddButton(info, level)
end

local old_ToggleDropDownMenu = ToggleDropDownMenu
ToggleDropDownMenu = function(level, value, dropDownFrame, anchorName, xOffset, yOffset)
	spacerDropDown = nil
	old_ToggleDropDownMenu(level, value, dropDownFrame, anchorName, xOffset, yOffset)
end

print = function(...)
	local s = ""
	for i = 1, table.getn(arg) do
		if s ~= "" then
			s = s .. " "
		end
		if type(arg[i]) == "table" and type(arg[i][0]) == "userdata" then
			local name = arg[i]:GetName()
			if name == "" or name == nil then
				name = ""
			else
				name = ":" .. name
			end
			s = s .. "<" .. arg[i]:GetObjectType() .. name .. ">"
		else
			s = s .. tostring(arg[i])
		end
	end
	if string.len(s) == 0 then
		s = "\n"
	end
	DEFAULT_CHAT_FRAME:AddMessage(s)
end

local start
tableToString = function(t, depth)
	if t[".recurse"] then
		return "<recursion>"
	end
	t[".recurse"] = 1
	if depth == nil then
		depth = 0
		start = GetTime()
	end
	if GetTime() - start >= 0.05 then
		return "{too large}"
	end
	local s = ""
	s = "{\n"
	for k,v in pairs(t) do
		if k ~= ".recurse" then
			s = s .. string.rep("    ", depth + 1)
			if type(k) == "table" then
				s = s .. "[" .. tableToString(k, depth + 1) .. "]"
			elseif type(k) == "string" then
				local u = format("[%q]", k)
				if not string.find(k, "[^A-Za-z0-9_]") and string.sub(u, 3, -3) == k then
					s = s .. k
				else
					s = s .. u
				end
			else
				s = s .. "[" .. tostring(k) .. "]"
			end
			s = s .. " = "
			if type(v) == "table" then
				s = s .. tableToString(v, depth + 1)
			elseif type(v) == "string" then
				s = s .. format("%q", v)
			else
				s = s .. tostring(v)
			end
			s = s .. ",\n"
		end
	end
	s = s .. string.rep("    ", depth) .. "}"
	t[".recurse"] = nil
	if depth == 0 then
		start = nil
	end
	return s
end
local tableToString = tableToString

printFull = function(k, t)
	if t == nil then
		k, t = nil, k
	end
	local s
	if type(t) == "table" and type(t[0]) == "userdata" then
		local name = t:GetName()
		if name == "" or name == nil then
			name = ""
		else
			name = ":" .. name
		end
		t = "<" .. t:GetObjectType() .. name .. ">"
	end
	if type(t) == "table" and type(t[0]) ~= "userdata" then
		if k == nil then
			s = tableToString(t)
		else
			s = k .. " = " .. tableToString(t)
		end
	else
		if k == nil then
			s = t
		else
			s = k .. " = " .. t
		end
	end
	string.gsub(s, "[^\n]+", print)
end

codeposition = function(num)
	return strsub(debugstack((num or 0) + 2, 1, 0), 0, -5)
end

SLASH_PRINT1 = "/print"
SlashCmdList["PRINT"] = function(text)
	RunScript(format("local func = function(...) if type(arg[1]) == 'table' then printFull(arg[1]) else print(unpack(arg)) end end;func(%s)", text))
end

SLASH_RELOAD1 = "/reload"
SLASH_RELOAD2 = "/rl"
SlashCmdList["RELOAD"] = function()
	ReloadUI()
end
