local private = private

local MinimapContainer = MinimapContainer
local PanelClass = PanelClass
local tablet = TabletLib:GetInstance('1.0')

local MINOR_VERSION = tonumber(string.sub("$Revision: 1643 $", 12, -3))
if MINOR_VERSION > FUBAR_REVISION then
	FUBAR_REVISION = MINOR_VERSION
end

local initialized = false
function MinimapContainer:Initialize()
	if not initialized then
		initialized = true
	else
		return
	end
	if FuBar.data.minimap == nil then
		FuBar.data.minimap = {}
	end
	
	self.plugins = {}
	self.data = {}
	self.data.plugins = FuBar.data.minimap
end

function MinimapContainer:AddPlugin(plugin)
	if private.changingProfile or plugin.userDefinedFrame then
		return
	end
	self:Initialize()
	if plugin.panel ~= nil then
		plugin.panel:RemovePlugin(plugin)
	end
	plugin.panel = self
	if not plugin.minimapFrame then
		local frame = CreateFrame("Button", plugin.frame:GetName() .. "MinimapButton", Minimap)
		plugin.minimapFrame = frame
		plugin:UpdateTooltip()
		tablet:Register(frame, plugin.frame)
		frame.plugin = plugin
		frame:SetWidth(31)
		frame:SetHeight(31)
		frame:SetFrameStrata("HIGH")
		frame:SetHighlightTexture("Interface\\Minimap\\UI-Minimap-ZoomButton-Highlight")
		local icon = frame:CreateTexture(frame:GetName() .. "Icon", "BACKGROUND")
		plugin.minimapIcon = icon
		local path = plugin:GetIcon() or "Interface\\Icons\\INV_Misc_QuestionMark"
		icon:SetTexture(path)
		if string.sub(path, 1, 16) == "Interface\\Icons\\" then
			icon:SetTexCoord(0.05, 0.95, 0.05, 0.95)
		else
			icon:SetTexCoord(0, 1, 0, 1)
		end
		icon:SetWidth(20)
		icon:SetHeight(20)
		icon:SetPoint("TOPLEFT", frame, "TOPLEFT", 7, -5)
		local overlay = frame:CreateTexture(frame:GetName() .. "Overlay","OVERLAY")
overlay:SetTexture("Interface\\Minimap\\MiniMap-TrackingBorder")
		overlay:SetWidth(53)
		overlay:SetHeight(53)
		overlay:SetPoint("TOPLEFT",frame,"TOPLEFT")
		frame:EnableMouse(true)
		frame:RegisterForClicks("LeftButtonUp")
		frame.plugin = plugin
		--frame:SetMovable(true)
		frame:SetScript("OnClick", function()
			if not this.dragged then
				FuBar:Plugin_OnClick(this.plugin)
			end
		end)
		frame:SetScript("OnDoubleClick", function()
			FuBar:Plugin_OnDoubleClick(this.plugin)
		end)
		frame:SetScript("OnMouseDown", function()
			this.dragged = false
			if arg1 == "LeftButton" and not IsShiftKeyDown() and not IsControlKeyDown() and not IsAltKeyDown() then
				HideDropDownMenu(1)
				FuBar:Plugin_OnMouseDown(this.plugin, arg1)
			elseif arg1 == "RightButton" and not IsShiftKeyDown() and not IsControlKeyDown() and not IsAltKeyDown() then
				FuBar:Plugin_OpenMenu(this.plugin)
			else
				HideDropDownMenu(1)
				FuBar:Plugin_OnMouseDown(this.plugin, arg1)
			end
			if this.plugin.OnClick or this.plugin.OnMouseDown or this.plugin.OnMouseUp or this.plugin.OnDoubleClick then
				if string.sub(this.plugin.minimapIcon:GetTexture(), 1, 16) == "Interface\\Icons\\" then
					this.plugin.minimapIcon:SetTexCoord(0.14, 0.86, 0.14, 0.86)
				else
					this.plugin.minimapIcon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
				end
			end
		end)
		frame:SetScript("OnMouseUp", function()
			if not this.dragged then
				FuBar:Plugin_OnMouseUp(this.plugin, arg1)
			end
			if string.sub(this.plugin.minimapIcon:GetTexture(), 1, 16) == "Interface\\Icons\\" then
				this.plugin.minimapIcon:SetTexCoord(0.05, 0.95, 0.05, 0.95)
			else
				this.plugin.minimapIcon:SetTexCoord(0, 1, 0, 1)
			end
		end)
		frame:RegisterForDrag("LeftButton")
		frame:SetScript("OnDragStart", self.OnDragStart)
		frame:SetScript("OnDragStop", self.OnDragStop)
	end
	plugin.frame:Hide()
	plugin.minimapFrame:Show()
	self:ReadjustLocation(plugin)
	table.insert(self.plugins, plugin)
	local exists = false
	for _,name in ipairs(self.data.plugins) do
		if name == plugin.realName then
			exists = true
		end
	end
	if not exists then
		table.insert(self.data.plugins, plugin.realName)
	end
	return true
end

function MinimapContainer:ReadjustLocation(plugin)
	local frame = plugin.minimapFrame
	local position = plugin.data.minimapPosition or plugin.defaultMinimapPosition or math.random(1, 360)
	local angle = math.rad(position or 0)
	local x = math.cos(angle) * 80
	local y = math.sin(angle) * 80
	frame:SetPoint("CENTER", Minimap, "CENTER", x, y)
end

function MinimapContainer:RemovePlugin(index)
	if private.changingProfile then
		return
	end
	self:Initialize()
	if type(index) == "table" then
		index = self:IndexOfPlugin(index)
		if not index then
			return
		end
	end
	local t, dt = self.plugins, self.data.plugins
	local plugin = t[index]
	assert(plugin.panel == self, "Plugin has improper panel field")
	plugin.panel = nil
	if not private.PanelClass_stopUpdates then
		plugin.frame:Hide()
		plugin.minimapFrame:Hide()
	end
	for i = table.getn(dt), 1, -1 do
		if dt[i] == plugin.realName then
			table.remove(dt, i)
		end
	end
	table.remove(t, index)
	return true
end

function MinimapContainer:GetPlugin(index)
	return self.plugins[index]
end

function MinimapContainer:GetNumPlugins()
	return table.getn(self.plugins)
end

function MinimapContainer:IndexOfPlugin(plugin)
	for i,p in ipairs(self.plugins) do
		if p == plugin then
			return i, "MINIMAP"
		end
	end
end

function MinimapContainer:HasPlugin(plugin)
	return self:IndexOfPlugin(plugin) ~= nil
end

function MinimapContainer:GetPluginSide(plugin)
	local index = self:IndexOfPlugin(plugin)
	assert(index, "Plugin not in panel")
	return "MINIMAP"
end

function MinimapContainer.OnDragStart()
	this.dragged = true
	this:LockHighlight()
	this:SetScript("OnUpdate", MinimapContainer.OnUpdate)
	if string.sub(this.plugin.minimapIcon:GetTexture(), 1, 16) == "Interface\\Icons\\" then
		this.plugin.minimapIcon:SetTexCoord(0.05, 0.95, 0.05, 0.95)
	else
		this.plugin.minimapIcon:SetTexCoord(0, 1, 0, 1)
	end
end

function MinimapContainer.OnDragStop()
	this:SetScript("OnUpdate", nil)
	this:UnlockHighlight()
end

function MinimapContainer.OnUpdate()
	local mx, my = Minimap:GetCenter()
	local px, py = FuBar:GetScaledCursorPosition()
	local lastPosition = this.plugin.data.minimapPosition
	local position = math.deg(math.atan2(py - my, px - mx))
	if position <= 0 then
		position = position + 360
	elseif position > 360 then
		position = position - 360
	end
	this.plugin.data.minimapPosition = position
	MinimapContainer:ReadjustLocation(this.plugin)
end

function MinimapContainer:GetSavedOrder()
	return self.data.plugins
end
